/*

    File: compile.cpp

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <stdarg.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

//#define HACK_HACK		//output info to main_console
//#define VERBOSE_COMPILER
//#define OUTPUT_COMPILER

#ifdef HACK_HACK
#include "../m_foundation/main.h"
#define dprint main_console->Print
#else
#define dprint printf
#endif

#include "compile.h"


#ifdef OUTPUT_COMPILER
#include <time.h>
#endif

// ************ language ************ 

//<exprlist>	-> <expr> {"," <expr>}
//constrain expr() types to the "types[size]" array 
void CCodeCompiler::exprlist(int *type, int size) {
#ifdef VERBOSE_COMPILER
	dprint("exprlist = {\n");
#endif

	int srcStart, dstStart, dstEnd;

	srcStart = dstStart = dstEnd = program->code.getIndex();

	bool first = true;
	do {
		//if we're not first on the scene then pass by the comma...
		if (!first) lex->process();
		first = false;

		if (!size) {
			error("I didn't expect any more params");
		}

		dstStart = dstEnd;

		//for this expr(), the call'd function will eat up the push'd contents
		int exprType = expr();
		if (exprType != *type) {
			typeError("exprlist", *type, exprType);
		}

		dstEnd = program->code.getIndex();

		if (!program->code.swapBlock(srcStart, dstStart, dstEnd)) {
			error("failed to swap code block");
		}

		type++;
		size--;

	} while (lex->token() == TOKEN_COMMA);

	if (size) {
		error("I expected %d more params", size);
	}
#ifdef VERBOSE_COMPILER
	dprint("}\n");
#endif
}

#ifdef ATTEMPT_IMPLEMENT_FORK

//stmt_fork
void CCodeCompiler::label_fork() {

	program->code.addFork(0);
	int jumpForEndSource = program->code.getIndex();

	int stmtStart = program->code.getIndex();

	if (lex->token() == TOKEN_BRACE_OPEN) {
		stmtblock();
	} else {
		stmt();
	}

	program->code.addExit();	//force exit thread by the new thread

	int stmtEnd = program->code.getIndex();

	//modify the fork param to tell the fork stmt where to jump to
	program->code.set(jumpForEndSource - 1, program->code.getIndex() - jumpForEndSource);
}

#endif

bool CCodeCompiler::label_var(bool write, rw_addr_t *rw) {

#ifdef VERBOSE_COMPILER
	dprint("label = {\n");
#endif

////begin rip for getting variable addresses
////
////incoming data: the lexeme parser must be looking at the name ...
////outgoing data: the address of the name, and whether it is relative or not, and the label type
////
////...or we could do it this way...
////
////incoming data: the lexeme parser must be looking at the name ...
////				whether to push (reading) or pop (writing) from the address
////outgoing data: label type

	//store the last lexeme...
	char lastlex[256];
	strncpy(lastlex, lex->lexemeString(), sizeof(lastlex));

	//read the next lexeme
	lex->process();

//#error messed up somewhere
//	cant think at the moment

	//zero the offset
	rw->offset = 0;

	//if we got no valid var
	int varIndex;
	if (!program->getVariableID(lastlex, &varIndex)) {

		if (write) {
			error("the variable %s has not been defined for writing", lastlex);
		} else {
			//then see if its a function ptr ref
			int funcIndex;
			if (!program->getFunctionID(lastlex, &funcIndex)) {	// >=0 => user, <0 => predef
				error("the variable %s has not been defined", lastlex);
				return false;
			} else {
				int funcType;
				if (!program->getFunctionType(funcIndex, &funcType)) {
					error("WHAT IS UP?! WE GOT A VALID FUNCTION %s INDEX %d WITH INVALID TYPE!\n", lastlex, funcIndex);
					return false;
				}
#ifdef VERBOSE_COMPILER
				dprint(" - is a function with index %d type %d\n", funcIndex, funcType);
#endif
				rw->write = false;
				rw->type = ~funcType;
				rw->constant = true;
				rw->value = funcIndex;
				//program->code.addPushConst(funcIndex);
			}
		}
	} else {
		rw->type = program->getVariableType(varIndex);


		//see if we got a "name.field..."
		//TODO -  move this back so we test for it before testing for a parenthesis
		//this way we can have struct fields used as function pointers
		char lastDot[256];
		strncpy(lastDot, lastlex, sizeof(lastDot));
		while ( !strcmp(lex->lexemeString(), ".") ) {
			if (rw->type < PT_COUNT) {
				error("tried to . an un-structed variable type");
				return false;
			}
			int structIndex = rw->type - PT_COUNT;

			//skip past the dot
			lex->process();

			//get the struct
			CStruct *structure = &(program->structTable[structIndex]);

			//copy over the name
			strncpy(lastDot, lex->lexemeString(), sizeof(lastDot));

			//skip past the name
			lex->process();
			
			//find the matching field
			int fieldIndex = structure->getFieldIndex(lastDot);
			if (fieldIndex == -1) {
				error("got an invalid named field: %s\n", lastDot);
				break;
			}

			//add by its offset
			rw->offset += structure->getFieldOffset(program, fieldIndex);

			//and recalc the type
			rw->type = structure->field[fieldIndex].typeIndex;
		}

		rw->constant = false;
		rw->relative = program->isVarRelative(varIndex);
		rw->write = write;
		
		//if we get an external var - 
		if (varIndex < 0) {
			//just write out the index
			rw->value = varIndex;

		//if not - internal (global or relative) -
		} else {
			//for relative we write out the address
			//but what about globals?  how do we get the address?
			rw->value = program->getVariableAddress(varIndex);
		}
	}

#ifdef VERBOSE_COMPILER
	dprint("}\n");
#endif

	return true;
}

int CCodeCompiler::complete_rw(rw_addr_t rw) {

#ifdef HACK_HACK
	dprint("adding complete_rw\n");
#endif

	if (rw.constant) {
		program->code.addPushConst(rw.value);
	} else {
		int typeSize = program->getTypeSize(rw.type);
		if (rw.relative) {
			int addr = rw.value + rw.offset;
			if (rw.write) {
				if (typeSize == 1) {
					program->code.addPopRel(addr);	//local variables are +1 
				} else {
					program->code.addPopStrRel(addr, typeSize);
				}
			} else {
				if (typeSize == 1) {
					program->code.addPushRel(addr);	//local variables are +1 
				} else {
					program->code.addPushStrRel(addr, typeSize);
				}
			}
		} else {

#ifdef HACK_HACK
			dprint("adding operation write:%d value:%d type %d typeSize:%d\n",
				rw.write, rw.value, rw.type, typeSize);
			dprint("current pos is line %d col %d\n", lex->lineCount, lex->colCount);
#endif

			if (rw.write) {
				if (typeSize == 1) {
					program->code.addPopGlob(rw.value);			//global variables - direct value
				} else {
					program->code.addPopStrGlob(rw.value, typeSize);
				}
			} else {
				if (typeSize == 1) {
					program->code.addPushGlob(rw.value);			//global variables - direct value
				} else {
					program->code.addPushStrGlob(rw.value, typeSize);
				}
			}
		}
	}

	return rw.type;
}

//retursn the type of the return type for the function
int CCodeCompiler::call_exprlist(int callTypeIndex) {

#ifdef DATA_STRING
	CFuncType *funcType = &(program->funcTypeTable[callTypeIndex]);
#else
	CFuncType *funcType = program->funcTypeTable + callTypeIndex;
#endif
	int labelType = funcType->returnType;
	lex->process();

	int returnSize = program->getTypeSize(labelType);
	//makeway for return info beforehand
	if (returnSize) {
		program->code.addAddSP( returnSize );
	}

	//TODO - this doesn't consider user's giving () when they NEED expressions
	if (lex->token() == TOKEN_PAR_CLOSE) {
		if (funcType->paramCount) {
			error("I expected some expressions");
		}
		lex->process();
	} else {
		exprlist(funcType->paramTypes, funcType->paramCount);
		if (lex->token() == TOKEN_PAR_CLOSE) {
			lex->process();
		} else {
			error("i expected a closing parenthesis, i got a %s", tokenName[lex->token()]);
		}
	}

	return labelType;
}

int CCodeCompiler::label_process_call(rw_addr_t rw) {
	int labelType = -1;

	//index of the function to use - if we do use a function
	int callTypeIndex = -1;
	if (rw.constant) {
		if (!program->getFunctionType(rw.value, &callTypeIndex)) {
			error("WE COULDNT FIND THE CONSTANT TYPE!\n");
			return -1;
		}
	} else {
		if (rw.type >= 0) {
			error("the variable is not declared as a function type");
			lex->process();
			return -1;
		} else {
			callTypeIndex = ~rw.type;
		}
	}
	if (callTypeIndex < 0) {
		error("tried to evaluate a function for a non-func type");
	} else {

		labelType = call_exprlist(callTypeIndex);

		//for kcisk
		CFuncType *funcType = &(program->funcTypeTable[callTypeIndex]);

		//add in our call code to the stack
		complete_rw(rw);

		program->code.addCall();

		//cleanup params afterwards
		int funcTypeParamSetTypeSize = funcType->getParamSetTypeSize(program);

		if (funcTypeParamSetTypeSize) {
			program->code.addAddSP( -funcTypeParamSetTypeSize );
		}
	}
	
	return labelType;
}

//<label>	-> name ["(" [<exprlist>] ")"]
//returns the type of the label generated
int CCodeCompiler::label() {

	int labelType = PT_INVALID;

#ifdef VERBOSE_COMPILER
	dprint("label = {\n");
#endif
	if (lex->token() == TOKEN_NAME) {

		//if we got a reserved constant (true / false)
		if (!strcmp(lex->lexemeString(), "false")) {
			lex->process();						//read the next lexeme
			program->code.addPushConst(0);
			labelType = PT_BOOL;
		} else if (!strcmp(lex->lexemeString(), "true")) {
			lex->process();						//read the next lexeme
			program->code.addPushConst(1);
			labelType = PT_BOOL;
#ifdef ATTEMPT_IMPLEMENT_FORK
		} else if (!strcmp(lex->lexemeString(), "_fork")) {
			lex->process();						//read the next lexeme
			lex->process();				//should both these be here?  no worries 'til we implement it again...
			label_fork();
			labelType = PT_THREAD;
#endif

		//if we got an unaccounted name...
		} else {

			//pull out the label info...
			rw_addr_t rw;
			if (!label_var(false, &rw)) {	//not writeable
				return -1;
			}

			if (lex->token() == TOKEN_PAR_OPEN) {

				//verifies types
				//processes exprlist
				//completes rw
				labelType = label_process_call(rw);
			} else {
				//store the return type from the rw
				labelType = complete_rw(rw);
			}
		//we got a variable
//		} else {
//			labelType = complete_rw(label_var(false));
		}
	} else if (lex->token() == TOKEN_CONST_INT) {
		program->code.addPushConst(atoi(lex->lexemeString()));
		lex->process();
		labelType = PT_INT;
	} else if (lex->token() == TOKEN_CONST_FLOAT) {
		float f = (float)atof(lex->lexemeString());
		int *i = (int *)(&f);
		program->code.addPushConst(*i);	//convert float->int
		lex->process();
		labelType = PT_FLOAT;
	} else if (lex->token() == TOKEN_CONST_STRING) {
		int strIndex = program->registerString(lex->lexemeString());
		program->code.addPushConst(strIndex);
		lex->process();
		labelType = PT_STRING;
	} else if (lex->token() == TOKEN_SEMICOLON) {	//i.e. nothing for this line (used in for() stmts)
		labelType = PT_VOID;		//DO NOT PROCESS - this is a blank line - wait for the stmt reader to pick it up (should this happen?)
	} else {
		error("expected a name!, i got a %s", tokenName[lex->token()]);
		lex->process();
		labelType = PT_VOID;
	}
#ifdef VERBOSE_COMPILER
	dprint("}\n");
#endif

	if (labelType == PT_INVALID) {
		error("somehow I am returning a PT_INVALID labelType!");
	}

	return labelType;
}

/*
order of op:

		1	(), []							12
		2	new
		3	.
		4	++, --
		5	- (unary), !, ~, ++, --			11
		6	*, /, %							10
		7	+, -							9
		8	<<, >>							8
		9	<, >, <=, >=					7
		10	==, !=							6
		11	&								5
		12	^								4
		13	|								3
		14	&&								2
		15	||								1
		16	?:
		17	=, *=, /=, %=, -=, <<=, >>=, >>>=, &=, ^=, |=

*/

int CCodeCompiler::exprPar() {
#ifdef VERBOSE_COMPILER
	dprint("exprPar = {\n");
#endif
	int exprType = PT_INVALID;
	if (lex->token() == TOKEN_PAR_OPEN) {
		lex->process();
		//no guarantee this expr() will get eaten - give factor() the duty
		exprType = expr();
		if (lex->token() != TOKEN_PAR_CLOSE) {
			error("I WANTED A CLOSING PARENTHESIS!");
		}
		lex->process();
	} else {
		exprType = label();
	}
#ifdef VERBOSE_COMPILER
	dprint("}\n");
#endif
	return exprType;
}

int CCodeCompiler::exprUnary() {
#ifdef VERBOSE_COMPILER
	dprint("exprUnary = {\n");
#endif
	int foundToken = TOKEN_UNKNOWN;

	if (lex->token() == TOKEN_SUB ||
		lex->token() == TOKEN_NOT_BOOL ||
		lex->token() == TOKEN_NOT_INT)
	{
		foundToken = lex->token();
		lex->process();
	}

	int exprType = exprPar();

	switch (foundToken) {
	case TOKEN_UNKNOWN:
		break;				//nothing prefix'd
	case TOKEN_SUB:			//negate int/float
		if (exprType == PT_INT) {
			program->code.addNegInt();
		} else if (exprType == PT_FLOAT) {
			program->code.addNegFloat();
		} else {
			error("tried to negate a non-int/float");
		}
		break;
	case TOKEN_NOT_BOOL:	//not bool
		if (exprType == PT_BOOL) {
			program->code.addNotBool();
		} else {
			error("! expects a proceeding bool type");
		}
		break;
	case TOKEN_NOT_INT:		//not int
		if (exprType == PT_INT) {
			program->code.addNotInt();
		} else {
			error("~ expects a proceeding int type");
		}
		break;
	}

#ifdef VERBOSE_COMPILER
	dprint("}\n");
#endif
	return exprType;
}

int CCodeCompiler::exprProd() {
#ifdef VERBOSE_COMPILER
	dprint("exprProd = {\n");
#endif
	int exprType = exprUnary();
	while(lex->token() == TOKEN_MUL || lex->token() == TOKEN_DIV || lex->token() == TOKEN_PERCENT) {
		int token = lex->token();
		lex->process();
		int nextType = exprUnary();

		if (exprType != nextType) {
			error("tried to mul/div/mod exprs of differing type");
		}

		if (exprType == PT_INT) {
			if (token == TOKEN_MUL) {
				program->code.addMulInt();
			} else if (token == TOKEN_DIV) {
				program->code.addDivInt();
			} else if (token == TOKEN_PERCENT) {
				program->code.addModInt();
			}
		} else if (exprType == PT_FLOAT) {
			if (token == TOKEN_MUL) {
				program->code.addMulFloat();
			} else if (token == TOKEN_DIV) {
				program->code.addDivFloat();
			} else if (token == TOKEN_PERCENT) {
				error("mod is undefined for floats");
			}
		} else {
			error("tried to mul/div/mod exprs of an undefined type");
		}
	}
#ifdef VERBOSE_COMPILER
	dprint("}\n");
#endif
	return exprType;
}

int CCodeCompiler::exprSum() {
#ifdef VERBOSE_COMPILER
	dprint("exprSum = {\n");
#endif
	int exprType = exprProd();
	while(lex->token() == TOKEN_ADD || lex->token() == TOKEN_SUB) {
		int token = lex->token();
		lex->process();
		int nextType = exprProd();

		if (exprType != nextType) {
			error("tried to add/sub exprs of differing type");
		}

		if (exprType == PT_INT) {
			if (token == TOKEN_ADD) {
				program->code.addAddInt();
			} else if (token == TOKEN_SUB) {
				program->code.addSubInt();
			}
		} else if (exprType == PT_FLOAT) {
			if (token == TOKEN_ADD) {
				program->code.addAddFloat();
			} else if (token == TOKEN_SUB) {
				program->code.addSubFloat();
			}
		} else {
			error("tried to add/sub exprs of an undefined type");
		}
	}
#ifdef VERBOSE_COMPILER
	dprint("}\n");
#endif
	return exprType;
}

int CCodeCompiler::exprShift() {
#ifdef VERBOSE_COMPILER
	dprint("exprShift = {\n");
#endif
	int exprType = exprSum();
	while(lex->token() == TOKEN_SHL || lex->token() == TOKEN_SHR) {
		int token = lex->token();
		lex->process();
		int nextType = exprSum();

		if (exprType != PT_INT) {
			error("shifting variables of non-int type");
		}

		if (token == TOKEN_SHL) {
			program->code.addSHLInt();
		} else if (token == TOKEN_SHR) {
			program->code.addSHRInt();
		}
	}
#ifdef VERBOSE_COMPILER
	dprint("}\n");
#endif
	return exprType;
}

int CCodeCompiler::exprIneq() {
#ifdef VERBOSE_COMPILER
	dprint("exprIneq = {\n");
#endif
	int exprType = exprShift();
	while(	lex->token() == TOKEN_G || lex->token() == TOKEN_L ||
			lex->token() == TOKEN_GE || lex->token() == TOKEN_LE)
	{
		int token = lex->token();
		lex->process();
		int nextType = exprShift();

		if (exprType != nextType) {
			error("comparing differing sizes");
		}

		switch (exprType) {
		case PT_INT:
			if (token == TOKEN_G) {
				program->code.addGInt();
			} else if (token == TOKEN_L) {
				program->code.addLInt();
			} else if (token == TOKEN_GE) {
				program->code.addGEInt();
			} else if (token == TOKEN_LE) {
				program->code.addLEInt();
			}
			break;
		case PT_FLOAT:
			if (token == TOKEN_G) {
				program->code.addGFloat();
			} else if (token == TOKEN_L) {
				program->code.addLFloat();
			} else if (token == TOKEN_GE) {
				program->code.addGEFloat();
			} else if (token == TOKEN_LE) {
				program->code.addLEFloat();
			}
			break;
		default:
			error("comparing variables of non-numeric type");
		}

		//convert type after performing operation
		exprType = PT_BOOL;
	}
#ifdef VERBOSE_COMPILER
	dprint("}\n");
#endif
	return exprType;
}

int CCodeCompiler::exprEquiv() {
#ifdef VERBOSE_COMPILER
	dprint("exprEquiv = {\n");
#endif
	int exprType = exprIneq();
	while(lex->token() == TOKEN_EQ || lex->token() == TOKEN_NE) {
		int token = lex->token();
		lex->process();
		int nextType = exprIneq();

		if (exprType != nextType) {
			error("comparing differing sizes");
		}

		if (program->getTypeSize(exprType) != 1) {
			error("comparing variables of non-unit size");
		}

		if (token == TOKEN_EQ) {
			program->code.addEQInt();
		} else if (token == TOKEN_NE) {
			program->code.addNEInt();
		}
		//convert type after performing operation
		exprType = PT_BOOL;
	}
#ifdef VERBOSE_COMPILER
	dprint("}\n");
#endif
	return exprType;
}

int CCodeCompiler::exprLogicAnd() {
#ifdef VERBOSE_COMPILER
	dprint("exprLogicAnd = {\n");
#endif
	int exprType = exprEquiv();
	while(lex->token() == TOKEN_AND_INT) {
		if (exprType != PT_INT) {
			error("expected an int before &");
		}
		lex->process();
		if (exprEquiv() != PT_INT) {
			error("expected an int after &");
		}
		program->code.addAndInt();
	}
#ifdef VERBOSE_COMPILER
	dprint("}\n");
#endif
	return exprType;
}


int CCodeCompiler::exprLogicXOr() {
#ifdef VERBOSE_COMPILER
	dprint("exprLogicXOr = {\n");
#endif
	int exprType = exprLogicAnd();
	while(lex->token() == TOKEN_XOR_INT) {
		if (exprType != PT_INT) {
			error("expected an int before ^");
		}
		lex->process();
		if (exprLogicAnd() != PT_INT) {
			error("expected an int after ^");
		}
		program->code.addXorInt();
	}
#ifdef VERBOSE_COMPILER
	dprint("}\n");
#endif
	return exprType;
}

int CCodeCompiler::exprLogicOr() {
#ifdef VERBOSE_COMPILER
	dprint("exprLogicOr = {\n");
#endif
	int exprType = exprLogicXOr();
	while(lex->token() == TOKEN_OR_INT) {
		if (exprType != PT_INT) {
			error("expected an int before |");
		}
		lex->process();
		if (exprLogicXOr() != PT_INT) {
			error("expected an int after |");
		}
		program->code.addOrInt();
	}
#ifdef VERBOSE_COMPILER
	dprint("}\n");
#endif
	return exprType;
}

int CCodeCompiler::exprAnd() {
#ifdef VERBOSE_COMPILER
	dprint("exprAnd = {\n");
#endif
	int exprType = exprLogicOr();
	while(lex->token() == TOKEN_AND_BOOL) {
		if (exprType != PT_BOOL) {
			error("expected a bool before &&");
		}
		lex->process();
		if (exprLogicOr() != PT_BOOL) {
			error("expected a bool after &&");
		}
		program->code.addAndInt();
	}
#ifdef VERBOSE_COMPILER
	dprint("}\n");
#endif
	return exprType;
}

int CCodeCompiler::exprOr() {
#ifdef VERBOSE_COMPILER
	dprint("exprOr = {\n");
#endif
	int exprType = exprAnd();
	while(lex->token() == TOKEN_OR_BOOL) {
		if (exprType != PT_BOOL) {
			error("expected a bool before ||");
		}
		lex->process();
		if (exprAnd() != PT_BOOL) {
			error("expected a bool after ||");
		}
		program->code.addOrInt();
	}
#ifdef VERBOSE_COMPILER
	dprint("}\n");
#endif
	return exprType;
}

int CCodeCompiler::exprCond() {
#ifdef VERBOSE_COMPILER
	dprint("exprCond = {\n");
#endif

	int exprType = exprOr();

	if (lex->token() == TOKEN_QUESTION) {

		if (exprType != PT_BOOL) {
			error("expected a bool before ?");
		}
		lex->process();

		program->code.addJumpZero(0);
		int ifJumpSource = program->code.getIndex();

		int exprType1 = exprOr();

		if (lex->token() != TOKEN_COLON) {
			error("expected a : after ?");
		}
		lex->process();

		//jump from the end of the '?' past the ':'
		program->code.addJumpZero(0);
		int elseJumpSource = program->code.getIndex();

		//write the '?' jump location
		program->code.set(ifJumpSource - 1, program->code.getIndex() - ifJumpSource);

		int exprType2 = exprOr();

		if (exprType1 != exprType2) {
			error("expression types within ?: do not match");
		}

		//copy over the new return type - since 'exprType' was used to grab our boolean
		exprType = exprType1;

		//write the 'else' jump location
		program->code.set(elseJumpSource - 1, program->code.getIndex() - elseJumpSource);

//		if (exprAnd() != PT_BOOL) {
//			error("expected a bool after ||");
//		}
//		program->code.addOrInt();
	}
#ifdef VERBOSE_COMPILER
	dprint("}\n");
#endif
	return exprType;
}


//<type>		-> typename ["(" [<type> {"," <type>} ] ")" ]
//returns the associated type.
//negative type <=> logical NOT the bits, then it is a reference in function type table
int CCodeCompiler::type(bool force) {
#ifdef VERBOSE_COMPILER
	dprint("type = {\n");
#endif
	int resultType = PT_INVALID;
	if (lex->token() == TOKEN_NAME) {
		//store the primitive type
		int primitiveType = program->getPrimOrStructTypeID(lex->lexemeString());
		//if we get a valid primitive type, then check to read more...
		if (primitiveType != PT_INVALID) {
			//and go to the next line
			lex->process();
			//if ... then read a function ... else read a variable
			if (lex->token() == TOKEN_PAR_OPEN) {

				CFuncType buildingFuncType;
				buildingFuncType.clear();
				buildingFuncType.returnType = primitiveType;

				lex->process();
				while (lex->token() != TOKEN_PAR_CLOSE) {
					int paramType = type(true);	//must find a type
					if (paramType == PT_INVALID) {
						error("expected a type, found %s", lex->lexemeString());
						lex->process();
						break;
					}
					buildingFuncType.addParam(paramType);
					//next - expect comma
					if (lex->token() != TOKEN_COMMA) {
						if (lex->token() == TOKEN_PAR_CLOSE) {
							break;	//good stuff =)
						}
						error("expected a comma, found %s", lex->lexemeString());
						lex->process();
						break;
					}
					//get past the comma
					lex->process();
				}
				lex->process();

				//logical NOT it since its a functype
				resultType = ~program->registerFuncType(buildingFuncType);
			} else {
				resultType = primitiveType;
			}
		} else if (force) {
			error("%s is not a valid typename", lex->lexemeString());
			lex->process();
		}
	} else if (force) {		//force a type read <=> compain if we read none
		error("expected a type name, found %s", lex->lexemeString());
		lex->process();
	}
#ifdef VERBOSE_COMPILER
	{
		char resultTypeName[64];
		strcpy(resultTypeName, program->getTypeName(resultType));
		dprint("returning type %d / %s\n", resultType, resultTypeName);
	}
	dprint("}\n");
#endif
	return resultType;
}

//FIX THIS UP TO SUPPORT RW'S
void CCodeCompiler::stmt_decl(int varType) {
	char varname[256];

	//expect a name
	if (lex->token() != TOKEN_NAME) {
		error("expected <variable type> <variable name>...");
	}

	//store the last name
	strncpy(varname, lex->lexemeString(), sizeof(varname));

	int index;
	if (program->getVariableID(varname, &index)) {
		error("variable \"%s\" has already been defined", varname);
		lex->process();
	}

	//interpret the name
	index = program->registerVariable(varname, varType);
	//and cycle next

	int relAddr = program->getVariableAddress(index);

	lex->process();
	//check for an assignment
	if (lex->token() == TOKEN_ASSIGN) {
		//cycle next
		lex->process();
		//get expr
		//expr here pops the right val
		int exprType = expr();

		if (exprType != varType) {
			typeError("stmt_decl", varType, exprType);
		} else {

			//EXACTLY WHAT WE FIND IN THE stmt_return() CODE!
			int typeSize = program->getTypeSize(exprType);
			if (typeSize== 1) {
				program->code.addPopRel( relAddr);
			} else {
				program->code.addPopStrRel( relAddr,  typeSize);
			}
		}
	}
}

void CCodeCompiler::stmt_return() {
	if (lex->token() == TOKEN_SEMICOLON) {	//see if we're returning nothing
		if (currentFuncType.returnType != PT_VOID) {	//double check we're supposed to return nothing
			error("expected a return value type %d", currentFuncType.returnType);
		}
	} else {
		int exprType = expr();		//move it to the 'return' area
		//test type
		if (exprType != currentFuncType.returnType) {
			typeError("stmt_return", currentFuncType.returnType, exprType);
		}

		//move it to return area
		int returnOffset =
			- 1											//pass the IP
			- currentFuncType.getParamSetTypeSize(program)		//pass the params
			- program->getTypeSize(currentFuncType.returnType);	//to the start of ret data

		//JUST LIKE stmt_decl()
		int returnTypeSize = program->getTypeSize(exprType);
dprint("RETURN TYPE SIZE: %d\n", returnTypeSize);
		if (returnTypeSize == 1) {
			program->code.addPopRel( returnOffset);
		} else {
			program->code.addPopStrRel( returnOffset,  returnTypeSize);
		}
	}
	//and return
	program->code.addRet();
}

void CCodeCompiler::stmt_if() {
	if (lex->token() != TOKEN_PAR_OPEN) {
		error("expected ( after \"if\"");
	}
	if (expr() != PT_BOOL) {
		error("expected a boolean expression in the \"if\"");
	}
	program->code.addJumpZero(0);
	int ifJumpSource = program->code.getIndex();
	if (lex->token() == TOKEN_BRACE_OPEN) {
		stmtblock();
	} else {
		stmt();
	}

	if (lex->token() == TOKEN_NAME && !strcmp(lex->lexemeString(), "else")) {
		lex->process();

		//jump from the end of the 'if' code past the 'else' code
		program->code.addJump(0);
		int elseJumpSource = program->code.getIndex();

		//write the 'if' jump location
		program->code.set(ifJumpSource - 1, program->code.getIndex() - ifJumpSource);

		//process the 'else' code
		if (lex->token() == TOKEN_BRACE_OPEN) {
			stmtblock();
		} else {
			stmt();
		}

		//write the 'else' jump location
		program->code.set(elseJumpSource - 1, program->code.getIndex() - elseJumpSource);
	} else {
		//jump straight to the end
		program->code.set(ifJumpSource - 1, program->code.getIndex() - ifJumpSource);
	}
}

void CCodeCompiler::stmt_while() {
	int whileLoopDest = program->code.getIndex();
	if (lex->token() != TOKEN_PAR_OPEN) {
		error("expected ( after \"while\"");
	}
	if (expr() != PT_BOOL) {
		error("expected a boolean expression in the \"while\"");
	}
	program->code.addJumpZero(0);
	int whileJumpSource = program->code.getIndex();
	if (lex->token() == TOKEN_BRACE_OPEN) {
		stmtblock();
	} else {
		stmt();
	}
	program->code.addJump(whileLoopDest - program->code.getIndex() - 2);	//sub 2 to accomodate for the jump op
	program->code.set(whileJumpSource-1, program->code.getIndex() - whileJumpSource);
}

/*
for (a;b;c) d

NAME PAR_OPEN stmt expr SEMICOLON expr PAR_CLOSE (stmt|stmt_block)

converts to

<a>
for_begin:
<b>
JZ for_end
<d>
<c>
JMP for_begin
for_end:
*/

void CCodeCompiler::stmt_for() {
	if (lex->token() != TOKEN_PAR_OPEN) {
		error("expected ( after \"for\"");
	}
	lex->process();

	stmt();

	int forBeginDest = program->code.getIndex();

	if (expr() != PT_BOOL) {
		error("for arg 2 expected boolean type");
	}
	program->code.addJumpZero(0);
	int jumpForEndSource = program->code.getIndex();

	if (lex->token() != TOKEN_SEMICOLON) {
		error("expected ; after \"for\"");
	}
	lex->process();

	int exprStart = program->code.getIndex();
	stmt();

	if (lex->token() != TOKEN_PAR_CLOSE) {
		error("expected ) after \"for\", found %s", lex->lexemeString());
	}
	lex->process();

	int stmtStart = program->code.getIndex();

	if (lex->token() == TOKEN_BRACE_OPEN) {
		stmtblock();
	} else {
		stmt();
	}

	int stmtEnd = program->code.getIndex();

	if (!program->code.swapBlock(exprStart, stmtStart, stmtEnd)) {
		error("failed to swap code block");
	}

	program->code.addJump(forBeginDest - program->code.getIndex() - 2);	//sub 2 for jump op

	program->code.set(jumpForEndSource - 1, program->code.getIndex() - jumpForEndSource);
}

//<stmt>		-> ([<type>] varname_id ["=" <expr>] | <expr>) ";"
void CCodeCompiler::stmt() {
#ifdef VERBOSE_COMPILER
	dprint("stmt = {\n");
#endif

	bool needSemicolon = true;

	//get the type - record it while we're at it
	int varType = type(false);

//dprint("stmt : found var type %d\n", varType);

	if (varType != PT_INVALID) {
		stmt_decl(varType);
	} else {
//		int varIndex = program->variableTableIndex(lex->lexemeString());
//		//if we got a var first then
//		if (varIndex != -1) {

//TODO - test for special words first - then test for variables type
		//if we found a valid variable...
		int varIndex;
		if (program->getVariableID(lex->lexemeString(), &varIndex)) {

//dprint("stmt : found var id %d\n", varIndex);

			//see if an equals sign is next
//			if (lex->token(1) == TOKEN_ASSIGN) { 

//dprint("stmt : found TOKEN_ASSIGN\n");


			//grab var w/pop *here*
			//TODO - instead - get the address & whether or not its relative or global
			//THEN - if we find an equals next, process the expr() and Pop() it
			// OR  - if we don't ... then we'd have to backtrack the lexeme parser, wouldnt we?
			//or maybe i'm looking at this all wrong...
			//
			//or can we work around the stmt_assign?
			rw_addr_t rw;
			if (!label_var(false, &rw)) {	//'false' implies we /cannot/ write - so we /can/ read function pointers
				error("label_var FAILED\n");
				return;
			}

			//notice, the write code most proceed the "var = expr" code
			//soo... we could do some code block copying...

			//assert that this lexeme is the equals...
			//lex->process();	//pass by the var name

			switch (lex->token()) {
			case TOKEN_ASSIGN:
				//handle assignment
				{

					//cycle next - over equals
					lex->process();
					//get expr
					int exprType = expr();
					if (exprType != rw.type) {
						typeError("stmt", rw.type, exprType);
					} else {
						rw.write = true;	//we are writing!
						varType = complete_rw(rw);
					}
				}
				break;
			case TOKEN_PAR_OPEN:
				//handle function call

//TODO - til this is implemented, function pointer calls wont work

				////hmm... eval the expr list, then execute the call
				{
					int exprType = label_process_call(rw);
					int exprTypeSize = program->getTypeSize(exprType);
					if (exprTypeSize) {
						program->code.addAddSP(-exprTypeSize);
					}
				}
				
				break;
			default:
				error("not a statement");
				break;
			}


		} else {

//dprint("stmt : testing for special words\n");

			//check for special words *HERE*
			if (!strcmp(lex->lexemeString(), "return")) {
				lex->process();		//pass by the 'return'
				stmt_return();
			} else if (!strcmp(lex->lexemeString(), "if")) {
				lex->process();
				stmt_if();
				needSemicolon = false;
			} else if (!strcmp(lex->lexemeString(), "while")) {
				lex->process();
				stmt_while();
				needSemicolon = false;
			} else if (!strcmp(lex->lexemeString(), "for")) {
				lex->process();
				stmt_for();
				needSemicolon = false;
			} else {	//process your ordinary expression...
				int exprType = expr();
				int exprTypeSize = program->getTypeSize(exprType);
				if (exprTypeSize) {
					program->code.addAddSP(-exprTypeSize);
				}
			}
		}
	}

	if (needSemicolon) {
		if (lex->token() != TOKEN_SEMICOLON) {
			error("expected ;, found %s", lex->lexemeString());
		}
		lex->process();
	}

#ifdef VERBOSE_COMPILER
	dprint("}\n");
#endif
}

//<stmtblock>	-> "{" {<stmt>} "}"
void CCodeCompiler::stmtblock() {
#ifdef VERBOSE_COMPILER
	dprint("stmtblock = {\n");
#endif
	if (lex->token() != TOKEN_BRACE_OPEN) {
		error("expected {, found %s", lex->lexemeString());
	} else {

		//scoping of local vars
#ifdef DATA_STRING
		int variableCount = program->variableTable.size();
#else
		int variableCount = program->variableTableCount;
#endif

		lex->process();
		while(1) {
			if (lex->token() == TOKEN_EOF) {
				error("expected } or stmt, found EOF");
				break;
			} else if (lex->token() == TOKEN_BRACE_CLOSE) {
				break;
			} else if (lex->token() == TOKEN_BRACE_OPEN) {
				stmtblock();
			} else {
				stmt();
			}
		}

		//scoping
#ifdef DATA_STRING
		program->variableTable.resize(variableCount);
#else
		program->variableTableCount = variableCount;
#endif
	}
	//skip next - either } or the bad one
	lex->process();

#ifdef VERBOSE_COMPILER
	dprint("}\n");
#endif
}

//<param>		-> <type> paramname
void CCodeCompiler::param() {
#ifdef VERBOSE_COMPILER
	dprint("param = {\n");
#endif

	int paramType = type(true);	//force
	if (paramType != PT_INVALID) {
		currentFuncType.addParam(paramType);
		//todo - add based upon 'type's return
		if (lex->token() != TOKEN_NAME) {
			error("I expected a parameter name, found %s", lex->lexemeString());
		} else {
			//register it as a param =)
			program->registerVariable(lex->lexemeString(), paramType);
		}
	} else {
		error("expected <type>, found %s", lex->lexemeString());
	}

	lex->process();

#ifdef VERBOSE_COMPILER
	dprint("}\n");
#endif
}

//<paramlist>	-> "(" [<param> {"," <param>}] ")"
void CCodeCompiler::paramlist() {
#ifdef VERBOSE_COMPILER
	dprint("paramlist = {\n");
#endif
	if (lex->token() != TOKEN_PAR_OPEN) {
		error("expected an opening parenthesis, found %s", lex->lexemeString());
	} else {
		lex->process();
		while (lex->token() == TOKEN_NAME) {
			//get params
			param();
			if (lex->token() != TOKEN_COMMA) { //no more params
				break;
			}
			//skip past comma
			lex->process();
		}
		if (lex->token() != TOKEN_PAR_CLOSE) {
			error("expected a closing parenthesis, found %s", lex->lexemeString());
		}
	}
	lex->process();
#ifdef VERBOSE_COMPILER
	dprint("}\n");
#endif
}


//<structure>	->	structname "{" {<type> <fieldName> ";"} "}"
void CCodeCompiler::structure() {
#ifdef VERBOSE_COMPILER
	dprint("structure = {\n");
#endif
	char structName[256];

	//expect a name
	if (lex->token() != TOKEN_NAME) {
		error("expected <structure name> ...");
	}

	//store the last name
	strncpy(structName, lex->lexemeString(), sizeof(structName));
	lex->process();

	if (lex->token() != TOKEN_BRACE_OPEN) {
		error("expected { after structure name");
	}
	lex->process();

	int structureID = program->registerStruct(structName);
	if (structureID == -1) {
		error("messed up when registering a structure");
	} else {
		CStruct *structure = &(program->structTable[structureID]);

		while (lex->token() != TOKEN_BRACE_CLOSE) {
			int fieldType = type(true);
			if (lex->token() != TOKEN_NAME) {
				error("expected <structure field name>");
			}
			
			CStructParam structParam;
			structParam.typeIndex = fieldType;
			structParam.name = lex->lexemeString();

			//TODO - make sure the field isn't repeated!
			structure->field.push_back(structParam);
			lex->process();

			if (lex->token() != TOKEN_SEMICOLON) {
				error("I expected a semicolon!");
				break;
			}
			lex->process();
		}
		lex->process();
	}

#ifdef VERBOSE_COMPILER
	dprint("}\n");
#endif
}

//<func>		-> <type> funcname <paramlist> <stmtblock>
void CCodeCompiler::func() {
#ifdef VERBOSE_COMPILER
	dprint("func = {\n");
#endif

	//attempt to read through the type - if it exists
	int returnType = type(true);

	if (returnType != PT_INVALID) {
		char thisname[256];

		//get the type - record it while we're at it

		//expect a name
		if (lex->token() != TOKEN_NAME) {
			error("expected <function type> <function name>...");
		}

		//store the last name
		strncpy(thisname, lex->lexemeString(), sizeof(thisname));
		//and cycle next
		lex->process();

		switch(lex->token()) {
		case TOKEN_SEMICOLON:

			//register a global variable

			{
				int index;
				if (program->getVariableID(thisname, &index)) {
					error("variable \"%s\" has already been defined", thisname);
					lex->process();
				}

				//interpret the name
				index = program->registerVariable(thisname, returnType);
				//and cycle next
			}

			lex->process();	//skip past the ';' sign

			break;

		//case TOKEN_COMMA: we add another variable of the same type after this
		//case TOKEN_ASSIGN; we assign this variable an initial value ... initial stack?

		case TOKEN_PAR_OPEN:

			{

				currentFuncType.clear();
				currentFuncType.returnType = returnType;	//todo - set to 'type's gathered info
				//tell the var table that this is where the func params start
				program->beginParamVars();
				//check for params - grab as we go to currentFuncType
				paramlist();
				//register the func type
				int func_type = program->registerFuncType(currentFuncType);
				//register the name the name
				int funcIndex = program->registerFunction(thisname, func_type);
				//record where in the code we are starting at
				program->functionTable[funcIndex].codeOffset = program->code.getIndex();
				//tell the var table to push its current size -
				//that any proceeding vars are for this function only
				program->beginFuncVars(funcIndex);
				//do the statements
				stmtblock();
				//see if we need a return...
				//the safe way - always add a return
				program->code.addRet();
				//tell the var table that we're done - and where to store the info
				program->endFuncVars();

			}
			break;
		default:
			error("expected a function or global variable, found %s", lex->lexemeString());
			lex->process();
			break;
		}

	} else {
		error("expected <type>, found %s", lex->lexemeString());
		lex->process();
	}
#ifdef VERBOSE_COMPILER
	dprint("}\n");
#endif
}

//<funcs>		-> {"struct" <structure> | <func>}
void CCodeCompiler::funcs() {
#ifdef VERBOSE_COMPILER
	dprint("funcs = {\n");
#endif
	while (lex->token() != TOKEN_EOF) {
		if (!strcmp(lex->lexemeString(), "struct")) {
			lex->process();
			structure();
		} else {
			func();
		}
	}
#ifdef VERBOSE_COMPILER
	dprint("}\n");
#endif
}


void CCodeCompiler::error(const char *_format, ...) {
	if (foundError) return;	//already got our error message stored

#ifdef DATA_STRING

	char buffer[256];
	_snprintf(buffer, sizeof(buffer), "line %d col %d - ", lex->lineCount, lex->colCount);

	char *dest = buffer + strlen(buffer);

	va_list ap;
	va_start(ap, _format);
	_vsnprintf(dest, sizeof(buffer) - (dest - buffer), _format, ap);
	va_end(ap);

	errorStr = "error: ";
	errorStr += buffer;

#ifdef OUTPUT_COMPILER
	dprint("error: %s\n", buffer);
#endif
#else
	sprintf(errorStr, "line %d col %d - ", lex->lineCount, lex->colCount);
	char *dest = errorStr + strlen(errorStr);

	va_list ap;
	va_start(ap, _format);
	vsprintf(dest, _format, ap);
	va_end(ap);

#ifdef OUTPUT_COMPILER
	dprint("error: %s\n", errorStr);
#endif
#endif

	foundError = true;
}

void CCodeCompiler::typeError(const char *msg, int rightType, int wrongType) {
	char right[64], wrong[64];
	strcpy(right, program->getTypeName(rightType));
	strcpy(wrong, program->getTypeName(wrongType));
	error("%s: expr type %d / %s should be %d / %s", msg, wrongType, wrong, rightType, right);
}

bool CCodeCompiler::compile() {

#ifdef OUTPUT_COMPILER
	clock_t startTime = clock();
#endif
	
	funcs();
	
#ifdef OUTPUT_COMPILER
	clock_t endTime = clock();

	program->print();

	dprint("completed in %f seconds\n", (double)(endTime - startTime) / (double)CLOCKS_PER_SEC);
#endif

	return !foundError;
}


CCodeCompiler::CCodeCompiler(CLexemeParser *lex, CProgram *program) {

	//init
	foundError = false;
#ifdef DATA_STRING
	errorStr = "";
#else
	memset(errorStr, 0, sizeof(errorStr));
#endif
	this->lex = lex;
	this->program = program;

	if (!lex) {
		error("CCodeCompiler - got a null lexeme parser");
		return;
	}
	if (!program) {
		error("CCodeCompiler - got a null program object");
		return;
	}
}


const char *CCodeCompiler::getError() {
#ifdef DATA_STRING
	return errorStr.c_str();
#else
	return errorStr ? errorStr : "";
#endif
}
