/*

    File: globalstack.cpp

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <assert.h>

#include "globalstack.h"

void CGlobalStack::clear(void) {
	size = 0;
	stack = NULL;
}

void CGlobalStack::kill(void) {
	if (stack) free(stack);
	clear();
}

CGlobalStack::CGlobalStack() {
	clear();
}

CGlobalStack::~CGlobalStack() {
	kill();
}

bool CGlobalStack::reset(CProgram *pr) {
	kill();
	assert(pr);
	size = pr->variableTableSize();	//read the required global stack size for this program
	stack = (int *)malloc(sizeof(stack[0]) * size);
	if (!stack) {
		size = 0;
		return false;
	}
	return true;
}

int CGlobalStack::get(int index) {
//	printf("CGlobalStack::get(%d)\n", index);
	if (index < 0 || index >= size) {
//		printf(" - size %d restriction\n", size);
		return -1;
	}
	return stack[index];
}

void CGlobalStack::set(int index, int data) {
	if (index < 0 || index >= size) return;
	stack[index] = data;
}
