/*

    File: interface.cpp

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <assert.h>

#include "interface.h"

#undef VERBOSE_INTERFACE
//#define VERBOSE_INTERFACE

CThreadStackInterface::CThreadStackInterface() {
	this->program = NULL;
	this->globalStack = NULL;
	this->dataStack = NULL;
	this->func = -1;
}

bool CThreadStackInterface::getParam(int param, void *result, int resultType) {
	if (!result) return false;
	if (!dataStack) return false;
	if (!program) return false;

	int resultSize = program->getTypeSize(resultType) * sizeof(int);	//times unit size

	//get function type associated with "thisFunc"
	if (func < 0 || func >= program->predefFunctionTable.size()) {
#ifdef VERBOSE_INTERFACE
		printf("interface error - current function index is invalid: %d of %d\n",
			func, program->predefFunctionTable.size());
#endif
		memset(result, -1, resultSize);
		return false;
	}
	CPredefFunction *function = &(program->predefFunctionTable[func]);
	int funcTypeIndex = function->functype;
	if (funcTypeIndex < 0 || funcTypeIndex >= program->funcTypeTable.size()) {
#ifdef VERBOSE_INTERFACE
		printf("interface error - current function's type index is invalid\n");
#endif
		memset(result, -1, resultSize);
		return false;
	}
	CFuncType *funcType = &(program->funcTypeTable[funcTypeIndex]);
	//find the "param"th parameter type
	if (param < 0 || param >= funcType->paramCount) {
#ifdef VERBOSE_INTERFACE
		printf("interface error - param index is invalid\n");
#endif
		memset(result, -1, resultSize);
		return false;
	}
	int paramType = funcType->paramTypes[param];
	if (resultType == -1) {	//any function
		if (paramType > 0) {
#ifdef VERBOSE_INTERFACE
			printf("interface error - param type is not a function type\n");
#endif
			memset(result, -1, resultSize);
			return false;
		}
	} else if (resultType >= PT_FIRST && resultType <= PT_LAST) {
		if (paramType != resultType) {
#ifdef VERBOSE_INTERFACE
			printf("interface error - param type is not desired type\n");
#endif
			memset(result, -1, resultSize);
			return false;
		}
	} else {
#ifdef VERBOSE_INTERFACE
		printf("interface error - desired type is invalid\n");
#endif
		memset(result, -1, resultSize);
		return false;
	}

	//calc the offset to this param
	int stackOffset = 0;	//start at the top of the stack
	for (int i = 0; i <= param; i++) {
		stackOffset += program->getTypeSize( funcType->paramTypes[param]);
	}
	assert(stackOffset > 0);

	//return the stack value at that point
	int stackSize = dataStack->size();
	if (stackSize < stackOffset) {
#ifdef VERBOSE_INTERFACE
		printf("interface error - stack offset underflows stack\n");
#endif
		memset(result, -1, resultSize);
		return false;
	}

#ifdef VERBOSE_INTERFACE
	printf("reading from stack offset %d of %d\n", stackSize - stackOffset, stackSize);
#endif
	memcpy(result, &(dataStack->at(stackSize - stackOffset)), resultSize);
#ifdef VERBOSE_INTERFACE
	printf("int value is %d\n", *(int *)result);
#endif
	return true;
}

bool CThreadStackInterface::getParamFunc(int param, int *result) {
	//-1 implies any function
	return getParam(param, result, -1);
}

bool CThreadStackInterface::getParamBool(int param, bool *result) {
	return getParam(param, result, PT_BOOL);
}

bool CThreadStackInterface::getParamInt(int param, int *result) {
	return getParam(param, result, PT_INT);
}

bool CThreadStackInterface::getParamFloat(int param, float *result) {
	return getParam(param, result, PT_FLOAT);
}

bool CThreadStackInterface::getParamEntity(int param, int *result) {
	return getParam(param, result, PT_ENTITY);
}

bool CThreadStackInterface::getParamThread(int param, int *result) {
	return getParam(param, result, PT_THREAD);
}

bool CThreadStackInterface::getParamString(int param, int *result) {
	return getParam(param, result, PT_STRING);
}

//#error THIS IS CAUSING FUNCTIONS TO CRASH!! FIX IT
bool CThreadStackInterface::setReturn(void *result, int resultType) {

#ifdef VERBOSE_INTERFACE
printf("testing object existance...\n");
#endif

	if (!result) return false;
	if (!dataStack) return false;
	if (!program) return false;

#ifdef VERBOSE_INTERFACE
printf("calculating result size...\n");
#endif

	int resultSize = program->getTypeSize(resultType) * sizeof(int);

#ifdef VERBOSE_INTERFACE
printf("testing predef function index validity...\n");
#endif

	//get function type associated with "thisFunc"
	if (func < 0 || func >= program->predefFunctionTable.size()) {
		//Error - current function index is invalid
		return false;
	}

#ifdef VERBOSE_INTERFACE
printf("getting predef function pointer...\n");
#endif

	CPredefFunction *function = &(program->predefFunctionTable[func]);

#ifdef VERBOSE_INTERFACE
printf("getting predef function type index...\n");
#endif

	int funcTypeIndex = function->functype;

#ifdef VERBOSE_INTERFACE
printf("getting predef function type index validity...\n");
#endif

	if (funcTypeIndex < 0 || funcTypeIndex >= program->funcTypeTable.size()) {
		//Error - current function's type index is invalid
		return false;
	}

#ifdef VERBOSE_INTERFACE
printf("getting predef function type...\n");
#endif

	CFuncType *funcType = &(program->funcTypeTable[funcTypeIndex]);

#ifdef VERBOSE_INTERFACE
printf("getting return type...\n");
#endif

	int returnType = funcType->returnType;

#ifdef VERBOSE_INTERFACE
printf("testing result type to return type...\n");
#endif

	if (resultType == -1) {	//any function
		if (returnType > 0) {
			//Error - param type is not a function type
			return false;
		}
	} else if (resultType >= PT_FIRST && resultType <= PT_LAST) {
		if (returnType != resultType) {
			//Error - param type is not desired type
			return false;
		}
	} else {
		//Error - desired type is invalid
		return false;
	}

#ifdef VERBOSE_INTERFACE
printf("calculating stack offset...\n");
#endif

	//calc the offset to this param
	int stackOffset = funcType->getParamSetTypeSize(program) + program->getTypeSize(funcType->returnType);	//to the start of ret data
	assert(stackOffset > 0);

#ifdef VERBOSE_INTERFACE
printf("getting stack size...\n");
#endif

	//return the stack value at that point
	int stackSize = dataStack->size();

#ifdef VERBOSE_INTERFACE
printf("comparing stack offset and size...\n");
#endif

	if (stackSize < stackOffset) {
		//Error - stack offset underflows stack
		return false;
	}

#ifdef VERBOSE_INTERFACE
printf("copying memory...\n");
#endif

	memcpy(&(dataStack->at(stackSize - stackOffset)), result, resultSize);

#ifdef VERBOSE_INTERFACE
printf("done...\n");
#endif

	return true;
}

bool CThreadStackInterface::setReturnFunc(int ret) {
	return setReturn(&ret, -1);	//-1 implies function
}
bool CThreadStackInterface::setReturnBool(bool ret) {
	return setReturn(&ret, PT_BOOL);
}
bool CThreadStackInterface::setReturnInt(int ret) {
	return setReturn(&ret, PT_INT);
}
bool CThreadStackInterface::setReturnFloat(float ret) {
	return setReturn(&ret, PT_FLOAT);
}
bool CThreadStackInterface::setReturnEntity(int ret) {
	return setReturn(&ret, PT_ENTITY);
}
bool CThreadStackInterface::setReturnThread(int ret) {
	return setReturn(&ret, PT_THREAD);
}
bool CThreadStackInterface::setReturnString(int ret) {
	return setReturn(&ret, PT_STRING);
}

const char *CThreadStackInterface::getString(int index) {
	if (program && index >= 0 && index < program->stringTable.size()) {
		return program->stringTable[index].c_str();
	}
	return "";
}
