/*

    File: lexeme.cpp

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <stdio.h>
#include <string.h>

#include "lexeme.h"

#define numberof(x) (sizeof(x)/sizeof(x[0]))

//#define HACK_HACK		//output info to main_console
//#define VERBOSE_LEXEME

#ifdef HACK_HACK
#include "../m_foundation/main.h"
#define dprint main_console->Print
#else
#define dprint printf
#endif

void CLexemeParser::clear(void) {
	memset(this, 0, sizeof(*this));
	lineCount = 1;
	colCount = 1;
	charClass = CHARCLASS_UNKNOWN;

	for (int i = 0; i < numberof(nextToken); i++) {
		nextToken[i] = TOKEN_UNKNOWN;
	}
}

CLexemeParser::CLexemeParser(void) {
	clear();
}

CLexemeParser::CLexemeParser(const char *src, int size) {
	clear();
	setReadBuffer(src, size);
}

void CLexemeParser::setReadBuffer(const char *src, int size) {

	if (this->readBuffer) {
#ifdef VERBOSE_LEXEME
		dprint("error - CLexemeParser::init - tried to init twice!\n");
#endif
		return;
	}

	clear();

	readBuffer = src;
	readBufferSize = size;
	readBufferIndex = 0;

	//now read the first char for the token parser
	getChar();	//start off the lex

	//and read the first token for the language parser
	for (int i = 0; i < LEXEME_DEPTH; i++) {
		process();		//should this be req?  yeah.  for read of subsequent calls
	}
}


//reads readBuffer->getChar(readBufferIndex)
//writes charClass, nextChar
void CLexemeParser::getChar(void) {
	
	colCount++;

	//read the next char
//	if (!readBuffer->inbounds(readBufferIndex)) {
	if (readBufferIndex < 0 || readBufferIndex >= readBufferSize) {
		nextChar = 0;
		charClass = CHARCLASS_EOF;		//special charclass_eof?
	}

	nextChar = readBuffer[readBufferIndex];
	readBufferIndex++;

	if ((nextChar >= 'a' && nextChar <= 'z')  ||
		(nextChar >= 'A' && nextChar <= 'Z') ||
		(nextChar == '_'))
	{
		charClass = CHARCLASS_LETTER;
	} else if (nextChar >= '0' && nextChar <= '9') {
		charClass = CHARCLASS_DIGIT;
	} else if (nextChar == ' ' || nextChar == '\t' || nextChar == '\r') {
		charClass = CHARCLASS_WHITESPACE;
	} else if (nextChar == '\n') {
		charClass = CHARCLASS_WHITESPACE;
		lineCount++;
		colCount = 1;
	} else if (nextChar == '\0') {
		charClass = CHARCLASS_EOF;	//why not? =)
	} else {
		charClass = CHARCLASS_UNKNOWN;
	}
}

//reads nextChar
//writes nextLexeme[lexemeIndex]
void CLexemeParser::addChar(void) {
	if (lastLexemeIndex < sizeof(lexeme[LEXEME_DEPTH-1])) {
		lexeme[LEXEME_DEPTH-1][lastLexemeIndex] = nextChar;
		lastLexemeIndex++;
	}
}

//reads - param lex
//returns - the token ident associated with 'lex' (for special words)
int CLexemeParser::lookup(const char *lex) {
	return TOKEN_NAME;
}

//returns 'true' if we're still inside the file
//returns false once we've stepped out of bounds
void CLexemeParser::process(void) {

	//copy over the old ones
	for (int i = 0; i < LEXEME_DEPTH-1; i++) {
		memcpy(lexeme[i], lexeme[i+1], sizeof(lexeme[0]));
		nextToken[i] = nextToken[i+1];
	}
	//zero the next - prep for reading
	memset(lexeme[LEXEME_DEPTH-1], 0, sizeof(lexeme[0]));
	nextToken[LEXEME_DEPTH-1] = TOKEN_UNKNOWN;

	//loop for when we find commented code
tryagain:

	//reset the index
	lastLexemeIndex = 0;

	//assume we've already getchar'd

	//automatically skip white spaces
	while (charClass == CHARCLASS_WHITESPACE) {
		getChar();
	}

	if (charClass == CHARCLASS_EOF) {
		nextToken[LEXEME_DEPTH-1] = TOKEN_EOF;
	} else if (charClass == CHARCLASS_LETTER) {
		addChar();
		getChar();
		while (charClass == CHARCLASS_LETTER || charClass == CHARCLASS_DIGIT) {
			addChar();
			getChar();
		}
		nextToken[LEXEME_DEPTH-1] = lookup(lexeme[LEXEME_DEPTH-1]);
	} else if (charClass == CHARCLASS_DIGIT || nextChar == '.') {
		bool foundFloat = false;
		foundFloat |= nextChar == '.';
		addChar();
		getChar();
		while (charClass == CHARCLASS_DIGIT || nextChar == '.') {
			foundFloat |= nextChar == '.';
			addChar();
			getChar();
		}
		//suffix 'f' for 'float'
		if (nextChar == 'f') {
			addChar();
			getChar();
			foundFloat = true;
		}
		nextToken[LEXEME_DEPTH-1] = foundFloat ? TOKEN_CONST_FLOAT : TOKEN_CONST_INT;
	} else if (nextChar == '\"') {
		getChar();
		while (nextChar != '\"') {
			if (charClass == CHARCLASS_EOF) {
#ifdef VERBOSE_LEXEME
				dprint("error - CLexemeParser::process(): found a broken string\n");
#endif
				//produce an error
				break;
			} else if (nextChar == '\\') {
				getChar();
//	\n New line
//	\r Carriage return
//	\t Horizontal tab
//	\v Vertical tab
//	\' Single quotation mark
//	\"  Double quotation mark
//	\\ Backslash
//	\? Literal question mark
//	\ooo ASCII character in octal notation
//	\xhhh ASCII character in hexadecimal notation
				switch (nextChar) {
				case 'a':		//not yet...
				case 'b':		//not yet...
				case 'f':		//not yet...
					nextChar = 12;
					addChar();
					getChar();
					break;
				case 'n':
					nextChar = 10;
					addChar();
					getChar();
					break;
				case 'r':
					nextChar = 13;
					addChar();
					getChar();
					break;
				case 't':
					nextChar = 9;
					addChar();
					getChar();
					break;
				case 'v':
					nextChar = 11;
					addChar();
					getChar();
					break;
//				case '\'':		//default:
//				case '\"':		//default:
//				case '\\':		//default:
//				case '?':		//default:
//				case 'o':		//not yet...
//				case 'x':		//not yet...
					getChar();
					break;
				default:
					addChar();
					getChar();
					break;
				}
			} else {
				addChar();
				getChar();
			}
		} 
		//pass over the last quote
		getChar();
		nextToken[LEXEME_DEPTH-1] = TOKEN_CONST_STRING;
	} else if (nextChar == '*') {
		addChar();
		getChar();
		nextToken[LEXEME_DEPTH-1] = TOKEN_MUL;
	} else if (nextChar == '/') {
		addChar();
		getChar();
		if (nextChar == '/') {		//read past comments
			while(nextChar != '\r' && nextChar != '\n') {
				getChar();
			}

			goto tryagain;
//			process();		//go at it again
//			return;			//and bail out
		} else if (nextChar == '*') {
			bool gotStarLast = false;
			do {
				getChar();
				if (nextChar == '*') {
					gotStarLast = true;
				} else {
					if (gotStarLast && nextChar == '/') {
						getChar();
						goto tryagain;
					}
					gotStarLast = false;	//clear the * buffer so that we always immediately need */
				}
			} while (1);
		} else {
			nextToken[LEXEME_DEPTH-1] = TOKEN_DIV;
		}
	} else if (nextChar == '+') {
		addChar();
		getChar();
		nextToken[LEXEME_DEPTH-1] = TOKEN_ADD;
	} else if (nextChar == '-') {
		addChar();
		getChar();
		nextToken[LEXEME_DEPTH-1] = TOKEN_SUB;
	} else if (nextChar == '(') {
		addChar();
		getChar();
		nextToken[LEXEME_DEPTH-1] = TOKEN_PAR_OPEN;
	} else if (nextChar == ')') {
		addChar();
		getChar();
		nextToken[LEXEME_DEPTH-1] = TOKEN_PAR_CLOSE;
	} else if (nextChar == ',') {
		addChar();
		getChar();
		nextToken[LEXEME_DEPTH-1] = TOKEN_COMMA;
	} else if (nextChar == '=') {
		addChar();
		getChar();
		if (nextChar == '=') {
			addChar();
			getChar();
			nextToken[LEXEME_DEPTH-1] = TOKEN_EQ;
		} else {
			nextToken[LEXEME_DEPTH-1] = TOKEN_ASSIGN;
		}
	} else if (nextChar == ';') {
		addChar();
		getChar();
		nextToken[LEXEME_DEPTH-1] = TOKEN_SEMICOLON;
	} else if (nextChar == '{') {
		addChar();
		getChar();
		nextToken[LEXEME_DEPTH-1] = TOKEN_BRACE_OPEN;
	} else if (nextChar == '}') {
		addChar();
		getChar();
		nextToken[LEXEME_DEPTH-1] = TOKEN_BRACE_CLOSE;
	} else if (nextChar == '[') {
		addChar();
		getChar();
		nextToken[LEXEME_DEPTH-1] = TOKEN_BRACKET_OPEN;
	} else if (nextChar == ']') {
		addChar();
		getChar();
		nextToken[LEXEME_DEPTH-1] = TOKEN_BRACKET_CLOSE;
	} else if (nextChar == '%') {
		addChar();
		getChar();
		nextToken[LEXEME_DEPTH-1] = TOKEN_PERCENT;
	} else if (nextChar == '!') {
		addChar();
		getChar();
		if (nextChar == '=') {
			addChar();
			getChar();
			nextToken[LEXEME_DEPTH-1] = TOKEN_NE;
		} else {
			nextToken[LEXEME_DEPTH-1] = TOKEN_NOT_BOOL;
		}

//NOTICE - THIS ISNT GETTING HIT THANKS TO OUR FLOAT PARSER
//SO INSTEAD OF SEARCHING FOR TOKEN_DOT, JUST SEARCH FOR THE STRING "."
	} else if (nextChar == '.') {
		addChar();
		getChar();
		nextToken[LEXEME_DEPTH-1] = TOKEN_DOT;
	} else if (nextChar == '~') {
		addChar();
		getChar();
		nextToken[LEXEME_DEPTH-1] = TOKEN_NOT_INT;
	} else if (nextChar == '<') {
		addChar();
		getChar();
		if (nextChar == '<') {
			addChar();
			getChar();
			nextToken[LEXEME_DEPTH-1] = TOKEN_SHL;
		} else if (nextChar == '=') {
			addChar();
			getChar();
			nextToken[LEXEME_DEPTH-1] = TOKEN_LE;
		} else {
			nextToken[LEXEME_DEPTH-1] = TOKEN_L;
		}
	} else if (nextChar == '>') {
		addChar();
		getChar();
		if (nextChar == '>') {
			addChar();
			getChar();
			nextToken[LEXEME_DEPTH-1] = TOKEN_SHR;
		} else if (nextChar == '=') {
			addChar();
			getChar();
			nextToken[LEXEME_DEPTH-1] = TOKEN_GE;
		} else {
			nextToken[LEXEME_DEPTH-1] = TOKEN_G;
		}
	} else if (nextChar == '&') {
		addChar();
		getChar();
		if (nextChar == '&') {
			addChar();
			getChar();
			nextToken[LEXEME_DEPTH-1] = TOKEN_AND_BOOL;
		} else {
			nextToken[LEXEME_DEPTH-1] = TOKEN_AND_INT;
		}
	} else if (nextChar == '|') {
		addChar();
		getChar();
		if (nextChar == '|') {
			addChar();
			getChar();
			nextToken[LEXEME_DEPTH-1] = TOKEN_OR_BOOL;
		} else {
			nextToken[LEXEME_DEPTH-1] = TOKEN_OR_INT;
		}
	} else if (nextChar == '^') {
		addChar();
		getChar();
		nextToken[LEXEME_DEPTH-1] = TOKEN_XOR_INT;
	} else if (nextChar == '?') {
		addChar();
		getChar();
		nextToken[LEXEME_DEPTH-1] = TOKEN_QUESTION;
	} else if (nextChar == ':') {
		addChar();
		getChar();
		nextToken[LEXEME_DEPTH-1] = TOKEN_COLON;
	} else {
#ifdef VERBOSE_LEXEME
		dprint ("error - bad token - nextChar %d \"%c\" charClass %d!\n", nextChar,nextChar,charClass);
#endif
		addChar();
		getChar();
		nextToken[LEXEME_DEPTH-1] = TOKEN_UNKNOWN;
	}

#ifdef VERBOSE_LEXEME
	dprint ("read token type %s - string \"%s\" - size %d - hex %x - line %d col %d\n",
		tokenName[nextToken[0]], lexeme[0], strlen(lexeme[0]), lexeme[0][0], lineCount, colCount);
#endif
}

int CLexemeParser::token(int index) {
	if (index < 0) index = 0;
	if (index >= LEXEME_DEPTH) index = LEXEME_DEPTH - 1;
	return nextToken[index];	//private-ize access - no writing from outside
}

const char *CLexemeParser::lexemeString(int index) {
	if (index < 0) index = 0;
	if (index >= LEXEME_DEPTH) index = LEXEME_DEPTH - 1;
#ifdef VERBOSE_LEXEME
	dprint("lexeme string: %s\n", lexeme[index]);
#endif
	return lexeme[index];
}
