/*

    File: lexeme.h

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#if !defined M_SCRIPT_LEXEME_H 
#define M_SCRIPT_LEXEME_H

enum {
	CHARCLASS_EOF,
	CHARCLASS_UNKNOWN,
	CHARCLASS_LETTER,
	CHARCLASS_DIGIT,
	CHARCLASS_WHITESPACE
};

enum {
	TOKEN_EOF,
	TOKEN_UNKNOWN,
	TOKEN_NAME,
	TOKEN_CONST_INT,
	TOKEN_CONST_FLOAT,
	TOKEN_CONST_STRING,
	TOKEN_MUL,
	TOKEN_DIV,
	TOKEN_ADD,
	TOKEN_SUB,
	TOKEN_PAR_OPEN,
	TOKEN_PAR_CLOSE,
	TOKEN_COMMA,
	TOKEN_ASSIGN,
	TOKEN_SEMICOLON,
	TOKEN_QUESTION,
	TOKEN_COLON,
	TOKEN_BRACE_OPEN,
	TOKEN_BRACE_CLOSE,
	TOKEN_BRACKET_OPEN,
	TOKEN_BRACKET_CLOSE,
	TOKEN_PERCENT,
	TOKEN_DOT,
	TOKEN_NOT_BOOL,
	TOKEN_NOT_INT,

	TOKEN_SHL,
	TOKEN_SHR,
	TOKEN_G,
	TOKEN_L,
	TOKEN_GE,
	TOKEN_LE,
	TOKEN_EQ,
	TOKEN_NE,
	TOKEN_AND_INT,
	TOKEN_OR_INT,
	TOKEN_XOR_INT,
	TOKEN_AND_BOOL,
	TOKEN_OR_BOOL
};

static const char *tokenName[] = {
	"TOKEN_EOF",
	"TOKEN_UNKNOWN",
	"TOKEN_NAME",
	"TOKEN_CONST_INT",
	"TOKEN_CONST_FLOAT",
	"TOKEN_CONST_STRING",
	"TOKEN_MUL",
	"TOKEN_DIV",
	"TOKEN_ADD",
	"TOKEN_SUB",
	"TOKEN_PAR_OPEN",
	"TOKEN_PAR_CLOSE",
	"TOKEN_COMMA",
	"TOKEN_ASSIGN",
	"TOKEN_SEMICOLON",
	"TOKEN_QUESTION",
	"TOKEN_COLON",
	"TOKEN_BRACE_OPEN",
	"TOKEN_BRACE_CLOSE",
	"TOKEN_BRACKET_OPEN",
	"TOKEN_BRACKET_CLOSE",
	"TOKEN_PERCENT",
	"TOKEN_NOT_BOOL",
	"TOKEN_NOT_INT",

	"TOKEN_SHL",
	"TOKEN_SHR",
	"TOKEN_G",
	"TOKEN_L",
	"TOKEN_GE",
	"TOKEN_LE",
	"TOKEN_EQ",
	"TOKEN_NE",
	"TOKEN_AND_INT",
	"TOKEN_OR_INT",
	"TOKEN_XOR_INT",
	"TOKEN_AND_BOOL",
	"TOKEN_OR_BOOL",
};

//TODO - make this defined per CLexemeParser instance
#define LEXEME_DEPTH	2

//make return characters optional to be read in?

class CLexemeParser {
//instance vars:
private:
	void clear(void);

	int charClass;
	char nextChar;

	char lexeme[LEXEME_DEPTH][256];
	int lastLexemeIndex;// = 0;
	int nextToken[LEXEME_DEPTH];	// = TOKEN_UNKNOWN;

	const char *readBuffer;
	int readBufferSize;
	int readBufferIndex;

public:
	int lineCount;
	int colCount;

//functions:
public:
	CLexemeParser();
	CLexemeParser(const char *src, int size);
	void setReadBuffer(const char *src, int size);
private:

	void getChar(void);
	void addChar(void);
	int lookup(const char *lex);

public:
	//gets the next lexeme (similar to getChar's operation)
	//stores in nextToken
	void process(void);
	//returns the currently stored lexeme
	int token(int index = 0);
	//returns the name
	const char *lexemeString(int index = 0);
	//returns the offset into the buffer
	int getBufferIndex(void) { return readBufferIndex; }
};


#endif