/*

    File: program.cpp

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

#include "program.h"

#define numberof(x)	(sizeof(x)/sizeof(x[0]))

#ifdef DATA_STRING
bool operator< (const CFuncType &a, const CFuncType &b) { return !memcmp(&a, &b, sizeof(CFuncType)); }
bool operator== (const CFuncType &a, const CFuncType &b) { return !memcmp(&a, &b, sizeof(CFuncType)); }
bool operator< (const CVariable &a, const CVariable &b) { return !memcmp(&a, &b, sizeof(CVariable)); }
bool operator== (const CVariable &a, const CVariable &b) { return !memcmp(&a, &b, sizeof(CVariable)); }
bool operator< (const CFunction &a, const CFunction &b) { return !memcmp(&a, &b, sizeof(CFunction)); }
bool operator== (const CFunction &a, const CFunction &b) { return !memcmp(&a, &b, sizeof(CFunction)); }
bool operator< (const CPredefVariable &a, const CPredefVariable &b) { return !memcmp(&a, &b, sizeof(CPredefVariable)); }
bool operator== (const CPredefVariable &a, const CPredefVariable &b) { return !memcmp(&a, &b, sizeof(CPredefVariable)); }
bool operator< (const CPredefFunction &a, const CPredefFunction &b) { return !memcmp(&a, &b, sizeof(CPredefFunction)); }
bool operator== (const CPredefFunction &a, const CPredefFunction &b) { return !memcmp(&a, &b, sizeof(CPredefFunction)); }
#endif

#undef VERBOSE_PROGRAM
//#define VERBOSE_PROGRAM

// ************ asm_gen ************ 

op_t ops[OP_COUNT] = {		//must align with the OP_ enum
	{"PUSH_CONST",	1},
	{"PUSH_REL",	1},
	{"PUSH_STR_REL",2},
	{"PUSH_GLOB",	1},
	{"PUSH_STR_GLOB",2},
	{"POP_REL",		1},
	{"POP_STR_REL",	2},
	{"POP_GLOB",	1},
	{"POP_STR_GLOB",2},
	{"ADD_SP",		1},
	{"JUMP",		1},
	{"JUMP_ZERO",	1},
#ifdef ATTEMPT_IMPLEMENT_FORK
	{"FORK",		1},
#endif
	{"CALL",		0},
	{"RET",			0},
	{"EXIT",		0},
	{"NOT_B",		0},
	{"MUL_I",		0},
	{"DIV_I",		0},
	{"MOD_I",		0},
	{"ADD_I",		0},
	{"SUB_I",		0},
	{"NEG_I",		0},
	{"NOT_I",		0},
	{"OP_SHL_I",	0},
	{"OP_SHR_I",	0},
	{"OP_G_I",		0},
	{"OP_L_I",		0},
	{"OP_GE_I",		0},
	{"OP_LE_I",		0},
	{"OP_EQ_I",		0},
	{"OP_NE_I",		0},
	{"OP_G_F",		0},
	{"OP_L_F",		0},
	{"OP_GE_F",		0},
	{"OP_LE_F",		0},
	{"OP_AND_I",	0},
	{"OP_XOR_I",	0},
	{"OP_OR_I",		0},
	{"MUL_F",		0},
	{"DIV_F",		0},
	{"ADD_F",		0},
	{"SUB_F",		0},
	{"NEG_F",		0},
};


void CCode::addPushConst(int val)		{last=OP_PUSH_CONST;push(last);push(val);			}
void CCode::addPushRel(int var_index)	{last=OP_PUSH_REL;	push(last);push(var_index);		}
void CCode::addPushStrRel(int addr, int size) {last=OP_PUSH_STR_REL; push(last); push(addr); push(size);}
void CCode::addPushGlob(int var_index)	{last=OP_PUSH_GLOB;	push(last);push(var_index);		}
void CCode::addPushStrGlob(int addr, int size) {last=OP_PUSH_STR_GLOB; push(last); push(addr); push(size);}
//void CCode::addPop()					{last=OP_POP;		push(last);						}
//void CCode::addPopStr(int value)		{last=OP_POP_STR;	push(last);push(value);			}
void CCode::addPopRel(int var_index)	{last=OP_POP_REL;	push(last);push(var_index);		}
void CCode::addPopStrRel(int addr, int size) {last=OP_POP_STR_REL; push(last); push(addr); push(size);}
void CCode::addPopGlob(int var_index)	{last=OP_POP_GLOB;	push(last);push(var_index);		}
void CCode::addPopStrGlob(int addr, int size) {last=OP_POP_STR_GLOB; push(last); push(addr); push(size);}
void CCode::addAddSP(int add)			{last=OP_ADD_SP;	push(last);push(add);			}
//void CCode::addPushBP()					{last=OP_PUSH_BP;	push(last);						}
//void CCode::addPopBP()					{last=OP_POP_BP;	push(last);						}
//void CCode::addSetBPtoSP()				{last=OP_SET_BP_TO_SP; push(last);					}
//void CCode::addSetSPtoBP()				{last=OP_SET_SP_TO_BP; push(last);					}
void CCode::addJump(int addr)			{last=OP_JUMP;		push(last);push(addr);}
void CCode::addJumpZero(int addr)		{last=OP_JUMP_ZERO;	push(last);push(addr);}
#ifdef ATTEMPT_IMPLEMENT_FORK
void CCode::addFork(int addr)			{last=OP_FORK;		push(last);push(addr);}
#endif
void CCode::addCall()					{last=OP_CALL;		push(last);}
void CCode::addRet()					{last=OP_RET;		push(last);}
void CCode::addExit()					{last=OP_EXIT;		push(last);}
void CCode::addNotBool()				{last=OP_NOT_B;		push(last);}
void CCode::addMulInt()					{last=OP_MUL_I;		push(last);}
void CCode::addDivInt()					{last=OP_DIV_I;		push(last);}
void CCode::addModInt()					{last=OP_MOD_I;		push(last);}
void CCode::addAddInt()					{last=OP_ADD_I;		push(last);}
void CCode::addSubInt()					{last=OP_SUB_I;		push(last);}
void CCode::addNegInt()					{last=OP_NEG_I;		push(last);}
void CCode::addNotInt()					{last=OP_NOT_I;		push(last);}
void CCode::addSHLInt()					{last=OP_SHL_I;		push(last);}
void CCode::addSHRInt()					{last=OP_SHR_I;		push(last);}
void CCode::addGInt()					{last=OP_G_I;		push(last);}
void CCode::addLInt()					{last=OP_L_I;		push(last);}
void CCode::addGEInt()					{last=OP_GE_I;		push(last);}
void CCode::addLEInt()					{last=OP_LE_I;		push(last);}
void CCode::addEQInt()					{last=OP_EQ_I;		push(last);}
void CCode::addNEInt()					{last=OP_NE_I;		push(last);}
void CCode::addGFloat()					{last=OP_G_F;		push(last);}
void CCode::addLFloat()					{last=OP_L_F;		push(last);}
void CCode::addGEFloat()				{last=OP_GE_F;		push(last);}
void CCode::addLEFloat()				{last=OP_LE_F;		push(last);}
void CCode::addAndInt()					{last=OP_AND_I;		push(last);}
void CCode::addXorInt()					{last=OP_XOR_I;		push(last);}
void CCode::addOrInt()					{last=OP_OR_I;		push(last);}
void CCode::addMulFloat()				{last=OP_MUL_F;		push(last);}
void CCode::addDivFloat()				{last=OP_DIV_F;		push(last);}
void CCode::addAddFloat()				{last=OP_ADD_F;		push(last);}
void CCode::addSubFloat()				{last=OP_SUB_F;		push(last);}
void CCode::addNegFloat()				{last=OP_NEG_F;		push(last);}

int CCode::getLastOp() { return last; }
int CCode::getIndex() {
#ifdef DATA_STRING
	return data.size();
#else
	return index;
#endif
}
int CCode::get(int i) {
#ifdef DATA_STRING
	return (i < 0 || i >= data.size()) ? OP_BAD : data[i];
#else
	return (i < 0 || i >= index) ? OP_BAD : data[i];
#endif
}
void CCode::set(int i, int val) {
#ifdef DATA_STRING
	data[i] = val;
#else
	if (i >= 0 && i < index) {
		data[i] = val;
	}
#endif
}
void CCode::push(int i) {
#ifdef DATA_STRING
	data.push_back(i);
#else
	if (index < numberof(data)) {
		data[index] = i;
		index++;
	}
#endif
}

int CCode::printOp(int index) {
	printf("%.8x - ", index);
	int opID = data[index];
	if (opID < 0 || opID >= numberof(ops)) {
		printf("invalid opID %d\n", opID);
		return index+1;
	}
	index++;
	printf("\t%s", ops[opID].name);
	for (int j = 0; j < ops[opID].param_count; j++) {
		printf("\t%.8x", data[index]);
		index++;
	}
	printf("\n");
	return index;
}

void CCode::print() {
#ifdef DATA_STRING
	for (int i = 0; i < data.size(); i = printOp(i));
	printf("code capacity: %d\n", data.capacity());
#else
	for (int i = 0; i < index; i = printOp(i));
#endif
}

bool CCode::swapBlock(int start, int mid, int end) {
#ifdef DATA_STRING


	int *temp = (int *)malloc((end - start) * sizeof(data[0]));
	if (!temp) {
		return false;
	} else {
		//memcpy(temp + (end - mid), &data[0] + start, (mid - start) * sizeof(data[0]));
		for (int j = 0; j < mid - start; j++) {
			temp[end - mid + j] = data[start + j];
		}
		//memcpy(temp, &data[0] + mid, (end - mid) * sizeof(data[0]));
		for (int j = 0; j < end - mid; j++) {
			temp[j] = data[mid + j];
		}
		//memcpy(&data[0] + start, temp, (end - start) * sizeof(data[0]));
		for (int j = 0; j < end - start; j++) {
			data[start + j] = temp[j];
		}
		free(temp);
	}


//	basic_string<int> b1 = data.substr(start, mid - start);
//	basic_string<int> b2 = data.substr(mid, end - mid);
//	data.replace(start + (end - mid), mid - start, b1);
//	data.replace(start, end - mid, b2);

#else
	if (end + (mid - start) < numberof(data)) {
		//reverse the order of ops added to the stack
		memcpy(data + end, data + start, (mid - start) * sizeof(data[0]));
		//C-style
		memcpy(data + start, data + mid, (end - start) * sizeof(data[0]));
	} else {
		int *temp = (int *)malloc((end - start) * sizeof(data[0]));
		if (!temp) {
			return false;
		} else {
			memcpy(temp + (end - mid), data + start, (mid - start) * sizeof(data[0]));
			memcpy(temp, data + mid, (end - mid) * sizeof(data[0]));
			memcpy(data + start, temp, (end - start) * sizeof(data[0]));
			free(temp);
		}
	}
#endif
	return true;
}

void CCode::clear() {
#ifdef DATA_STRING
//	data.erase(data.begin());
	data.clear();
#else
	index = 0;
#endif
}

CCode::CCode() {
	clear();
}


// ************ primitives ************ 

typedef struct primitive_s {
	const char *name;
	int dwsize;	//size, in DWORDs
} primitive_t;

primitive_t primitives[] = {
	{"",		0},	//invalid
	{"void",	0},
	{"bool",	1},
	{"int",		1},
	{"float",	1},
	{"entity",	1},
	{"thread",	1},
	{"string",	1},
};

int CProgram::getPrimTypeID(const char *str) {
	for (int i = PT_FIRST; i <= PT_LAST; i++) {
		if (!strcmp(str, primitives[i].name)) {
			return i;
		}
	}
	return PT_INVALID;
}

int CProgram::getTypeSize(int type) {
	if (type < 0) return 1;			//for all func ref's
	if (type == 0) {
#ifdef VERBOSE_PROGRAM
		printf("getTypeSize() - requested for type INVALID\n");
#endif
		return 0;		//INVALID
	}
	if (type >= PT_FIRST && type <= PT_LAST) {
		return primitives[type].dwsize;
	}

	//struct
	type -= PT_COUNT;
	if (type < 0 || type >= structTable.size()) {
#ifdef VERBOSE_PROGRAM
		printf("getTypeSize() - requested an invalid struct #%d\n", type);
#endif
		return 0;
	}
	return structTable[type].size(this);
}

const char *getPrimitiveTypeName(int type) {
	return (type >= PT_FIRST && type <= PT_LAST) ? primitives[type].name : "<invalid>";
}

//relies upon the function type table
const char *CProgram::getTypeName(int type) {
	static char strbuf[256];
	if (type < 0) {
		funcTypeTable[~type].getName(strbuf, sizeof(strbuf), this);
		return strbuf;
	} else if (type == 0) {
		return "<invalid>";
	} else if (type >= PT_FIRST && type <= PT_LAST) {
		return getPrimitiveTypeName(type);
	} else if (type - PT_COUNT < structTable.size()) {
		strncpy(strbuf, structTable[type - PT_COUNT].name.c_str(), sizeof(strbuf));
		return strbuf;
	}
	return "<invalid>";
}

// ************ structures ************ 

int CStruct::size(CProgram *program) {
	int sz = 0;
	for (int i = 0; i < field.size(); i++) {
		sz += program->getTypeSize(field[i].typeIndex);
	}
	return sz ? sz : 1;	//at least get one byte in there (so we dont have overlapping addrs)
}

int CStruct::getFieldIndex(const char *name) {
	for (int i = 0; i < field.size(); i++) {
		if (!strcmp(name, field[i].name.c_str())) return i;
	}
	return -1;
}

int CStruct::getFieldOffset(CProgram *program, int fieldIndex) {
	int ofs = 0;
	for (int i = 0; i < fieldIndex; i++) {
		ofs += program->getTypeSize(field[i].typeIndex);
	}
	return ofs;
}

int CProgram::getStructByName(const char *name) {
	for (int i = 0; i < structTable.size(); i++) {
		if (!strcmp(name, structTable[i].name.c_str())) {
			return i;
		}
	}
	return -1;
}

int CProgram::registerStruct(const char *name) {

	//make sure its not a primitive name <TODO - test for ALL KEYWORDS>
	for (int i = PT_FIRST; i <= PT_LAST; i++) {
		if (!strcmp(name, getPrimitiveTypeName(i))) {
			return -1;
		}
	}

	int pre = getStructByName(name);
	if (pre != -1) return pre;

	CStruct newStruct;
	newStruct.name = name;

	structTable.push_back(newStruct);
	return structTable.size()-1;
}

//returns 
int CProgram::getPrimOrStructTypeID(const char *name) {
	//test primitive
	int prim = getPrimTypeID(name);
	if (prim != PT_INVALID) return prim;

	//test structs
	int str = getStructByName(name);
	if (str != -1) return str + PT_COUNT;	//offset our structure index by 'count'

	//got nuttin'
	return PT_INVALID;
}

void CProgram::printStructTable(void) {
	for (int i = 0; i < structTable.size(); i++) {
		cout << "structure #" << i << " = \"" << structTable[i].name << "\"" << endl;
	}
}

// ************ function types ************ 

CFuncType::CFuncType() {
	clear();
}

CFuncType::CFuncType(int ret, int count, int *params) {
	clear();
	returnType = ret;

	while(count--) {
		addParam(*params);
		params++;
	}
}

void CFuncType::addParam(int type) {	//index ref to param table 
	if (paramCount > numberof(paramTypes)) {
#ifdef VERBOSE_PROGRAM
		printf("CFuncType::addParam() - filled up!\n");
#endif
		return;
	}
	paramTypes[paramCount] = type;
	paramCount++;
}

void CFuncType::clear() {
	memset(this, 0, sizeof(*this));
}

void CFuncType::printNumeric() {
	printf("%d : {", returnType);
	for (int i = 0; i < paramCount; i++) {
		if (i > 0) printf(", ");
		printf("%d", paramTypes[i]);
	}
	printf("}");
}

int CFuncType::getParamSetTypeSize(CProgram *program) {
	int size = 0;
	for (int i = 0; i < paramCount; i++) {
		size += program->getTypeSize(paramTypes[i]);
	}
	return size;
}

bool CFuncType::equals(const CFuncType *funcType) {
	if (returnType != funcType->returnType) return false;
	if (paramCount != funcType->paramCount) return false;
	return !memcmp(paramTypes, funcType->paramTypes, sizeof(int) * paramCount);
}

int CProgram::registerFuncType(CFuncType funcType) {

	//see if we already have a match...
#ifdef DATA_STRING
	for (int i = 0; i < funcTypeTable.size(); i++) {
		if (funcTypeTable[i].equals(&funcType)) return i;
	}

	funcTypeTable.push_back(funcType);
	return funcTypeTable.size()-1;

#else
	for (int i = 0; i < funcTypeTableCount; i++) {
		if (funcTypeTable[i].equals(&funcType)) return i;
	}

	if (funcTypeTableCount >= numberof(funcTypeTable)) {

#ifdef VERBOSE_PROGRAM
		printf("error - funcTypeTable overflow\n");
#endif
		return -1;
	}

	funcTypeTable[funcTypeTableCount] = funcType;
	funcTypeTableCount++;
	return funcTypeTableCount-1;
#endif
}

// ... all the way down here ...
//			no \n
void CFuncType::print(CProgram *program) {

	printf("%s (", program->getTypeName(returnType));

	if (paramCount) {
		for (int j = 0; j < paramCount; j++) {
			if (j > 0) {
				printf(", ");
			}
			int paramType = paramTypes[j];

			printf("%s", program->getTypeName(paramType));
		}
	}
	printf(")");
}

int CFuncType::getName(char *dest, int size, CProgram *program) {
	char *start = dest;
	if (returnType < 0) {
		int dif = program->funcTypeTable[~returnType].getName(dest, size, program);
		dest += dif;
		size -= dif;

		if (size > 2) {
			strcpy(dest, " (");
			dest += 2;
			size -= 2;
		} else {
			return (dest - start);
		}
	} else {
		const char *src = getPrimitiveTypeName(returnType);
		int srcSize = strlen(src);
		if (size > srcSize) {
			strcpy(dest, src);
			dest += srcSize;
			size -= srcSize;

			if (size > 2) {
				strcpy(dest, " (");
				dest += 2;
				size -= 2;
			} else {
				return (dest - start);
			}
		} else {
			return (dest - start);
		}
	}

	if (paramCount) {
		for (int j = 0; j < paramCount; j++) {
			if (j > 0) {
				if (size > 2) {
					strcpy(dest, ", ");
					dest += 2;
					size -= 2;
				} else {
					return (dest - start);
				}
			}

			int paramType = paramTypes[j];
			if (paramType < 0) {
				int dif = program->funcTypeTable[~paramType].getName(dest, size, program);
				dest += dif;
				size -= dif;
			} else {
				const char *src = getPrimitiveTypeName(paramType);
				int srcSize = strlen(src);
				if (size > srcSize) {
					strcpy(dest, src);
					dest += srcSize;
					size -= srcSize;
				} else {
					return dest - start;
				}
			}
		}
	}

	if (size > 1) {
		strcpy(dest, ")");
		dest++;
		size--;
	}
	return dest - start;
}

void CProgram::printFuncTypeTable(void) {
	int i = 0;
#ifdef DATA_STRING
	for (; i < funcTypeTable.size(); i++) {
		cout << "function type #" << i << " = ";
		funcTypeTable[i].printNumeric();
		printf(" = ");
		funcTypeTable[i].print(this);
		printf("\n");
	}
#else
	CFuncType *f = funcTypeTable;
	for (; i < funcTypeTableCount; i++, f++) {
		printf("function type #%d = ", i);
		f->printNumeric();
		printf(" = ");
		f->print(this);
		printf("\n");
	}
#endif
}

// ************ local variables ************ 

//something between funcs & vars
static int funcParam_variableTableCount = -1;
static int funcLocal_variableTableCount = -1;
static int funcLocal_funcIndex = -1;

int CVariable::size(CProgram *program) {
	return program->getTypeSize(type);
}

int CProgram::registerVariable(const char *name, int type) {

#ifdef DATA_STRING

	CVariable v;
	v.name = name;
	v.type = type;
	variableTable.push_back(v);

#else

	if (variableTableCount >= numberof(variableTable)) {
#ifdef VERBOSE_PROGRAM
		printf("error - variableTable overflow\n");
#endif
		return -1;
	}
	CVariable *v = variableTable + variableTableCount;
	variableTableCount++;
	strncpy(v->name, name, sizeof(v->name));
	v->type = type;
#endif

	//if this var belongs to a local function...
	if (funcLocal_funcIndex != -1) {
		//then set the new local size to the max of itself and the new local size
		CFunction *f = &(functionTable[funcLocal_funcIndex]);
		int localSize = 0;
#ifdef DATA_STRING
		for (int i = funcLocal_variableTableCount; i < variableTable.size(); i++) {
#else
		for (int i = funcLocal_variableTableCount; i < variableTableCount; i++) {
#endif
			localSize += variableTable[i].size(this);	//in DWORDs
		}
		if (localSize > f->localSize) {
			f->localSize = localSize;
		}
	}

#ifdef DATA_STRING
	return variableTable.size() - 1;
#else
	return v - variableTable;
#endif
}

int CProgram::variableTableIndex(const char *name) {
	//last to first =)
#ifdef DATA_STRING
	
	for (int i = variableTable.size() - 1; i >= 0; i--) {
		CVariable *v = &(variableTable[i]);
		if (!v->name.compare(name)) {
			return i;
		}
	}
#else
	int i = variableTableCount-1;
	CVariable *v = variableTable + i;
	for (; i >= 0; i--, v--) {
		if (!strncmp(name, v->name, sizeof(v->name))) {
			return i;
		}
	}
#endif
	return -1;
}

#ifndef SEPARATE_GLOBALS
int CProgram::variableTableSize(void) {
	int size = 0;
#ifdef DATA_STRING
	for (int i = variableTable.size()-1; i >= 0; i--) {
		CVariable *v = &(variableTable[i]);
#else
	for (CVariable *v = variableTable + variableTableCount - 1; v >= variableTable ; v--) {
#endif
		size += v->size(this);
	}
	return size;
}
#endif

void CProgram::printVariableTable(void) {
#ifdef DATA_STRING
	for (int i = 0; i < variableTable.size(); i++) {
		CVariable *v = &(variableTable[i]);
		cout << "variable #" << i << " = " << v->name << " type " << v->type << " : " << getTypeName(v->type) << endl;
	}
#else
	int i = 0;
	CVariable *v = variableTable;
	for (; i < variableTableCount; i++, v++) {
		printf("variable #%d = %s type %d : %s\n", i, v->name, v->type, getTypeName(v->type));
	}
#endif
}

bool CProgram::isVarRelative(int index) {
	return (index >= funcParam_variableTableCount);
}

// ************ all variables ************ 

//in goes the index, out goes the address
int CProgram::getVariableAddress(int index) {

	if (funcParam_variableTableCount == -1) {
#ifdef VERBOSE_PROGRAM
		printf("error - getVarRelativeAddr - no params have been defiend yet!\n");
		printf("FIXME - if no params the only dont support params - dont crash\n");
#endif
		return 0;
	}
	if (funcLocal_variableTableCount == -1) {
#ifdef VERBOSE_PROGRAM
		printf("error - getVarRelativeAddr - no locals have been defiend yet!\n");
		printf("FIXME - if no locals the only dont support locals - dont crash\n");
#endif
		return 0;
	}

	//if locals have been defined and we're on the + side of the locals
	//then return as a local

	//local
	if (index >= funcLocal_variableTableCount) {
		int addr = 1;
		//start at variable table entry and read up til we hit our var
		for (int i = funcLocal_variableTableCount; i < index; i++) {
			addr += getTypeSize(variableTable[i].type);
		}
		return addr;
	}

	//param
	if (index >= funcParam_variableTableCount) {
		int addr = -1;
#ifdef VERBOSE_PROGRAM
		printf("begin param addr comp\n");
#endif
		//start at the param table start and read up til we come to (including) our param
		for (int i = funcParam_variableTableCount; i <= index; i++) {
			addr -= getTypeSize(variableTable[i].type);
#ifdef VERBOSE_PROGRAM
			printf("offsetting by %d\n", getTypeSize(variableTable[i].type));
#endif
		}
#ifdef VERBOSE_PROGRAM
		printf("final addr: %x\n", addr);
#endif
		return addr;
	}

	//global
	if (index > 0) {
		if (index > variableTable.size()) {
#ifdef VERBOSE_PROGRAM
			printf("error - getVariableAddress - got an out of bound external variable index!\n");
#endif
			return 0;
		}
		int addr = 0;
		for (int i = 0; i < index; i++) {
			addr += getTypeSize(variableTable[i].type);
		}
		return addr;
	}

	//FALSE! RETURN FALSE! CHANGE THIS FUNCTION AROUND!
	return 0;
}

#ifdef SEPARATE_GLOBALS
// ************ global variables ************ 

int CProgram::registerGlobVar(const char *name, int type) {

	CVariable v;
	v.name = name;
	v.type = type;
	globVarTable.push_back(v);

	return variableTable.size() - 1;
}

int CProgram::globVarTableIndex(const char *name) {
	//last to first =)
	
	for (int i = globVarTable.size() - 1; i >= 0; i--) {
		CVariable *v = &(globVarTable[i]);
		if (!v->name.compare(name)) {
			return i;
		}
	}

	return -1;
}

//returns the size of the variable table stack.
//used for initializing threads to make room for the global variables
int CProgram::globVarTableSize(void) {
	int size = 0;
	for (int i = globVarTable.size()-1; i >= 0; i--) {
		CVariable *v = &(globVarTable[i]);
		size += v->size();
	}
	return size;
}

void CProgram::printGlobVarTable(void) {
	for (int i = 0; i < globVarTable.size(); i++) {
		CVariable *v = &(globVarTable[i]);
		cout << "glob var #" << i << " = " << v->name << " type " << v->type << " : " << getTypeName(v->type) << endl;
	}
}

#endif

// ************ predefined variables ************ 

int CProgram::registerPredefVariable(const char *name, int type, void *addr) {
#ifdef DATA_STRING

	CPredefVariable v;
	v.name = name;
	v.type = type;
	v.addr = (int *)addr;
	predefVariableTable.push_back(v);
	return predefVariableTable.size()-1;

#else

	if (predefVariableTableCount >= numberof(predefVariableTable)) {
#ifdef VERBOSE_PROGRAM
		printf("error - predefVariableTable overflow\n");
#endif
		return -1;
	}
	CPredefVariable *f = predefVariableTable + predefVariableTableCount;
	predefVariableTableCount++;
	strncpy(f->name, name, sizeof(f->name));
	f->type = type;
	f->addr = (int *)addr;
	return f - predefVariableTable;
#endif
}

int CProgram::predefVariableTableIndex(const char *name) {
#ifdef DATA_STRING
	for (int i = 0; i < predefVariableTable.size(); i++) {
		CPredefVariable *f = &(predefVariableTable[i]);
		if (!f->name.compare(name)) {
			return i;
		}
	}
	return -1;
#else
	int i = 0;
	CPredefVariable *f = predefVariableTable;
	for (; i < predefVariableTableCount; i++, f++) {
		if (!strncmp(name, f->name, sizeof(f->name))) {
			return i;
		}
	}
	return -1;
#endif
}


void CProgram::printPredefVariableTable(void) {
#ifdef DATA_STRING

	for (int i = 0; i < predefVariableTable.size(); i++) {
		CPredefVariable *f = &(predefVariableTable[i]);
		cout << "predef variable #" << i << " = " << f->name << "\ttype " << f->type << " : " << getTypeName(f->type) << endl;
	}

#else

	int i = 0;
	CPredefVariable *f = predefVariableTable;
	for (; i < predefVariableTableCount; i++, f++) {
		printf("predef variable #%d = %s\ttype %d : %s\n",
			i, f->name, f->type, getTypeName(f->type));
	}
#endif
}

// ************ functions ************ 

/*
index - placement in function table - not stored within function table, mind you =)
func_type - index of function type, stored in func_type table
name - index to string of name within the string table
local_var_count - size in the stack we should set aside for local vars?

index	func_type	name	parent_index	local_var_count

*/


int CProgram::registerFunction(const char *name, int functype) {

#ifdef DATA_STRING

	CFunction f;
	f.name = name;
	f.functype = functype;
	f.localSize = 0;
	functionTable.push_back(f);
	return functionTable.size()-1;

#else

	if (functionTableCount >= numberof(functionTable)) {
#ifdef VERBOSE_PROGRAM
		printf("error - functionTable overflow\n");
#endif
		return -1;
	}
	CFunction *f = functionTable + functionTableCount;
	functionTableCount++;
	strncpy(f->name, name, sizeof(f->name));
	f->functype = functype;
	f->localSize = 0;
	return f - functionTable;

#endif
}

int CProgram::functionTableIndex(const char *name) {
#ifdef DATA_STRING
	for (int i = 0; i < functionTable.size(); i++) {
		CFunction *f = &(functionTable[i]);
		if (!f->name.compare(name)) {
			return i;
		}
	}
#else
	int i = 0;
	CFunction *f = functionTable;
	for (; i < functionTableCount; i++, f++) {
		if (!strncmp(name, f->name, sizeof(f->name))) {
			return i;
		}
	}
#endif
	return -1;
}

void CProgram::printFunctionTable(void) {
#ifdef DATA_STRING

	for (int i = 0; i < functionTable.size(); i++) {
		CFunction *f = &(functionTable[i]);
		cout << "function #" << i << " = " << f->name << "\tlocalSize " << f->localSize
			<< "\tcodeOfs " << f->codeOffset << "\tfunctype " << f->functype << " : ";
		funcTypeTable[f->functype].print(this);
		printf("\n");
	}

#else
	int i = 0;
	CFunction *f = functionTable;
	for (; i < functionTableCount; i++, f++) {
		printf("function #%d = %s\tlocalSize %d\tcodeOfs %d\tfunctype %d : ",
			i, f->name, f->localSize, f->codeOffset, f->functype);
		funcTypeTable[f->functype].print(this);
		printf("\n");
	}
#endif
}

//call before interpreting function definition parameters
void CProgram::beginParamVars() {
	if (funcParam_variableTableCount != -1) {
#ifdef VERBOSE_PROGRAM
		printf("YOU tried to beginFuncVars() when you were already processing a function!\n");
#endif
		return;
	}
#ifdef DATA_STRING
	funcParam_variableTableCount = variableTable.size();
#else
	funcParam_variableTableCount = variableTableCount;
#endif
#ifdef VERBOSE_PROGRAM
	printf("beginning params at %d\n", funcParam_variableTableCount);
#endif
}

//call before interpreting function stmtblock
void CProgram::beginFuncVars(int funcIndex) {
	if (funcLocal_variableTableCount != -1) {
#ifdef VERBOSE_PROGRAM
		printf("YOU tried to beginFuncVars() when you were already processing a function!\n");
#endif
		return;
	}
#ifdef DATA_STRING
	if (funcIndex < 0 || funcIndex >= functionTable.size()) {
#else
	if (funcIndex < 0 || funcIndex >= functionTableCount) {
#endif
#ifdef VERBOSE_PROGRAM
		printf("beginFuncVars got a bad function index %d\n", funcIndex);
#endif
		return;
	}

#ifdef DATA_STRING
	funcLocal_variableTableCount = variableTable.size();
#else
	funcLocal_variableTableCount = variableTableCount;
#endif
	funcLocal_funcIndex = funcIndex;
#ifdef VERBOSE_PROGRAM
	printf("beginning locals for %d at %d\n", funcIndex, funcLocal_variableTableCount);
#endif

	//zero the local size.  increment it as we register variables
	functionTable[funcLocal_funcIndex].localSize = 0;
}

//call after function stmtblock
void CProgram::endFuncVars() {
	//restore - all the way back before the params
#ifdef DATA_STRING
	variableTable.resize(funcParam_variableTableCount);
#else
	variableTableCount = funcParam_variableTableCount;
#endif
	//reset
	funcLocal_variableTableCount = -1;
	funcParam_variableTableCount = -1;
	funcLocal_funcIndex = -1;
}

// ************ predefined functions ************ 

#include "thread.h"


int CProgram::registerPredefFunction(const char *name, int functype, void (*func)(CThreadStackInterface *)) {
#ifdef DATA_STRING

	CPredefFunction f;
	f.name = name;
	f.functype = functype;
	f.func = func;
	predefFunctionTable.push_back(f);
	return predefFunctionTable.size()-1;

#else

	if (predefFunctionTableCount >= numberof(predefFunctionTable)) {
#ifdef VERBOSE_PROGRAM
		printf("error - predefFunctionTable overflow\n");
#endif
		return -1;
	}
	CPredefFunction *f = predefFunctionTable + predefFunctionTableCount;
	predefFunctionTableCount++;
	strncpy(f->name, name, sizeof(f->name));
	f->functype = functype;
	f->func = func;
	return f - predefFunctionTable;
#endif
}

int CProgram::predefFunctionTableIndex(const char *name) {
#ifdef DATA_STRING
	for (int i = 0; i < predefFunctionTable.size(); i++) {
		CPredefFunction *f = &(predefFunctionTable[i]);
		if (!f->name.compare(name)) {
			return i;
		}
	}
#else
	int i = 0;
	CPredefFunction *f = predefFunctionTable;
	for (; i < predefFunctionTableCount; i++, f++) {
		if (!strncmp(name, f->name, sizeof(f->name))) {
			return i;
		}
	}
#endif
	return -1;
}


void CProgram::printPredefFunctionTable(void) {
#ifdef DATA_STRING

	
	for (int i = 0; i < predefFunctionTable.size(); i++) {
		CPredefFunction *f = &(predefFunctionTable[i]);
		cout << "predef function #" << i << " = " << f->name << "\tfunctype " << f->functype << " : ";
		funcTypeTable[f->functype].print(this);
		printf("\n");
	}

#else
	int i = 0;
	CPredefFunction *f = predefFunctionTable;
	for (; i < predefFunctionTableCount; i++, f++) {
		printf("predef function #%d = %s\tfunctype %d : ",
			i, f->name, f->functype);
		funcTypeTable[f->functype].print(this);
		printf("\n");
	}
#endif
}

// ************ the unified function set ************ 

bool CProgram::getFunctionID(const char *name, int *ret) {
	*ret = functionTableIndex(name);
	if (*ret != -1) return true;	//if we found one then bail

	*ret = predefFunctionTableIndex(name);
	if (*ret != -1) {
		*ret = ~(*ret);
		return true;
	}

	//no luck
	return false;
}

bool CProgram::getFunctionType(int index, int *ret) {
	if (index < 0) {
		index = ~index;
		if (index >= predefFunctionTable.size()) {
			return false;
		}
		*ret = predefFunctionTable[index].functype;
		return true;
	}

	if (index >= functionTable.size()) {
		return false;
	}
	*ret = functionTable[index].functype;
	return true;
}

// ************ the unified variable set ************ 

bool CProgram::getVariableID(const char *name, int *ret) {

	//if we find a local/param/global variable then return its index
	*ret = variableTableIndex(name);
	if (*ret != -1) return true;

	//if we find a predef variable then return its index, NOT'd
	*ret = predefVariableTableIndex(name);
	if (*ret != -1) {
		*ret = ~(*ret);
		return true;
	}

	//no luck
	return false;
}

int CProgram::getVariableType(int index) {
	if (index < 0) {
		return predefVariableTable[~index].type;
	} else {
		return variableTable[index].type;
	}	
}

// ************ strings ************ 

int CProgram::registerString(const char *name) {

#ifdef DATA_STRING

	stringTable.push_back(string(name));
	return stringTable.size()-1;

#else

	if (stringTableCount >= numberof(stringTable)) {
#ifdef VERBOSE_PROGRAM
		printf("error - stringTable overflow\n");
#endif
		return -1;
	}
	stringTable[stringTableCount] = name;
	stringTableCount++;
	return stringTableCount-1;

#endif
}

int CProgram::stringTableIndex(const char *name) {
#ifdef DATA_STRING
	for (int i = 0; i < stringTable.size(); i++) {
		if (!stringTable[i].compare(name)) {
			return i;
		}
	}
	return -1;
#else
	int i = 0;
	for (; i < stringTableCount; i++) {
		if (!stringTable[i].compare(name)) {
			return i;
		}
	}
	return -1;
#endif
}

void CProgram::printStringTable(void) {
#ifdef DATA_STRING

	for (int i = 0; i < stringTable.size(); i++) {
		cout << "string #" << i << " = \"" << stringTable[i] << "\"" << endl;
	}

#else

	for (int i = 0; i < stringTableCount; i++) {
		cout << "string #" << i << " = \"" << stringTable[i] << "\"" << endl;
	}

#endif
}

void CProgram::print() {
	printStructTable();
	printFuncTypeTable();
	printFunctionTable();
	printPredefFunctionTable();
	printVariableTable();
#ifdef SEPARATE_GLOBALS
	printGlobVarTable();
#endif
	printPredefVariableTable();
	printStringTable();
	code.print();
}

void CProgram::clear() {
	//reset it all
#ifdef DATA_STRING
	stringTable.clear();
	predefFunctionTable.clear();
	functionTable.clear();
	predefVariableTable.clear();
	variableTable.clear();
	funcTypeTable.clear();
	structTable.clear();
#else
	stringTableCount = 0;
	predefFunctionTableCount = 0;
	functionTableCount = 0;
	predefVariableTableCount = 0;
	variableTableCount = 0;
	funcTypeTableCount = 0;
#endif

	code.clear();
}


CProgram::CProgram() {
	clear();
}