/*

    File: program.h

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#if !defined M_SCRIPT_PROGRAM_H
#define M_SCRIPT_PROGRAM_H

class CProgram;	//mess

#include "interface.h"

//this determines where or not we use the stl stuffs
#define DATA_STRING

//#define ATTEMPT_IMPLEMENT_FORK

//the oldschool method - locals & globals all in the same address space
//#define SEPARATE_GLOBALS

#ifdef DATA_STRING
#include <vector>
#include <iostream>
#include <string>
using namespace std;
#endif

class CScriptThread;

// ************ code ************ 

enum {
	OP_BAD = -1,
	OP_PUSH_CONST,	//PUSH_CONST	<int>					- push immediate onto stack
	OP_PUSH_REL,	//PUSH_REL		<addr>					- push data at stack location [BP+addr]
	OP_PUSH_STR_REL,//PUSH_STR_REL	<addr>					- push <int> dwords from stack location [BP+addr]
	OP_PUSH_GLOB,	//PUSH_GLOB		<addr>					- push data at stack location [addr]
	OP_PUSH_STR_GLOB,//PUSH_STR_GLOB <addr> <int>			- push <int> dwords from stack location [addr]
//	OP_POP,			//POP									- pops into never never land
//	OP_POP_STR,		//POP_STR		<int>					- pops <int> dwords from the stack and throws them away
	OP_POP_REL,		//POP_REL		<addr>					- pops data into stack location [BP+addr]
	OP_POP_STR_REL,	//POP_STR_REL	<addr> <int>			= pops <int> dwords from the stack into stack location [BP+addr]
	OP_POP_GLOB,	//POP_GLOB		<addr>					- pops data into stack location [addr]
	OP_POP_STR_GLOB,//POP_STR_GLOB	<addr> <int>			- pops <int> dwords from the stack into stack location [addr]
	OP_ADD_SP,		//ADD_SP		<add>					= adds <add> to the SP register
//	OP_PUSH_BP,		//PUSH_BP								= push's BP onto the stack
//	OP_POP_BP,		//POP_BP								= pop's BP's contents
//	OP_SET_BP_TO_SP,//SET_BP_TO_SP							= sets BP to SP's contents
//	OP_SET_SP_TO_BP,//SET_SP_TO_BP							= sets SP to BP's contents
	OP_JUMP,		//JUMP			<addr>					= sets IP to IP+<addr>
	OP_JUMP_ZERO,	//JUMP_ZERO		<addr>					= pops data from stack.  if its zero then set IP to IP+<addr>
#ifdef ATTEMPT_IMPLEMENT_FORK
	OP_FORK,		//FORK			<addr>					= creates a new thread at current execution point.  set old thread IP to IP+<addr> 
#endif
	OP_CALL,		//CALL									- pops the dest function ID, pushs IP+sizeof(call op) and jumps to the function indexed by the int.  if int>=0 then calls a code function.  if int<0 then calls a predef function
	OP_RET,			//RET									- pops IP and jumps to IP
	OP_EXIT,		//EXIT									- termine execution
	OP_NOT_B,		//NOT_B									- pops an int.  if the val is zero then pushs 1 else pushes 0
	OP_MUL_I,		//MUL_I									- pops two ints from the stack, multiplies them, pushs the result back
	OP_DIV_I,		//DIV_I									- pops two ints from the stack, divides them, pushs the result back
	OP_MOD_I,		//MOD_I									- pops two ints from the stack, modulo's them, pushs the result back
	OP_ADD_I,		//ADD_I									- pops two ints from the stack, adds them, pushs the result back
	OP_SUB_I,		//SUB_I									- pops two ints from the stack, subtracts them, pushs the result back
	OP_NEG_I,		//NEG_I									- pops an int, negates it, pushs it
	OP_NOT_I,		//NOT_I									- pops an int, bitwise NOT's it, pushs it
	OP_SHL_I,		//SHL_I									- pops two ints, shifts the first to the left by the second, pushs the result back
	OP_SHR_I,		//SHR_I									- pops two ints, shifts the first to the right by the second, pushs the result back
	OP_G_I,			//G_I									- pops two ints, pushs 1 if the first is greater than the second, 0 otherwise
	OP_L_I,			//L_I									- pops two ints, pushs 1 if the first is less than the second, 0 otherwise
	OP_GE_I,		//GE_I									- pops two ints, pushs 1 if the first is greater than or equal to the second, 0 otherwise
	OP_LE_I,		//LE_I									- pops two ints, pushs 1 if the first is less than or equal to the second, 0 otherwise
	OP_EQ_I,		//EQ_I									- pops two ints, pushs 1 if the first is equal to the second, 0 otherwise
	OP_NE_I,		//NE_I									- pops two ints, pushs 1 if the first is not equal to the second, 0 otherwise
	OP_G_F,			//G_F									- pops two floats, pushs 1 if the first is greater than the second, 0 otherwise
	OP_L_F,			//L_F									- pops two floats, pushs 1 if the first is less than the second, 0 otherwise
	OP_GE_F,		//GE_F									- pops two floats, pushs 1 if the first is greater than or equal to the second, 0 otherwise
	OP_LE_F,		//LE_F									- pops two floats, pushs 1 if the first is less than or equal to the second, 0 otherwise
	OP_AND_I,		//AND_I									- pops two ints, bitwise AND's them, pushs the result back
	OP_XOR_I,		//XOR_I									- pops two ints, bitwise XOR's them, pushs the result back
	OP_OR_I,		//OR_I									- pops two ints, bitwise OR's them, pushs the result back
	OP_MUL_F,		//MUL_F									- pops two floats from the stack, multiplies them, pushs the result back
	OP_DIV_F,		//DIV_F									- pops two floats from the stack, divides them, pushs the result back
	OP_ADD_F,		//ADD_F									- pops two floats from the stack, adds them, pushs the result back
	OP_SUB_F,		//SUB_F									- pops two floats from the stack, subtracts them, pushs the result back
	OP_NEG_F,		//NEG_F									- pops a float, negates it, pushs it
	OP_COUNT
};

class CCode {
private:
//public:

#ifdef DATA_STRING
	vector<int> data;	//whats the difference between basic_string<int> and vector<int> ?
#else
	int index;
	int data[2048];
#endif

	int last;	//last op

public:

	void addPushConst(int val);
	//now i see why RISC is so popular...
	void addPushRel(int var_index);
	void addPushStrRel(int addr, int size);
	void addPushGlob(int var_index);
	void addPushStrGlob(int addr, int size);
//	void addPop();				//replaced with 'ADD_SP
//	void addPopStr(int value);	//replaced with 'ADD_SP
	void addPopRel(int var_index);
	void addPopStrRel(int addr, int size);
	void addPopGlob(int var_index);
	void addPopStrGlob(int addr, int size);
	void addAddSP(int add);
//	void addPushBP();			//all handled internally
//	void addPopBP();
//	void addSetBPtoSP();
//	void addSetSPtoBP();
	void addJump(int addr);
	void addJumpZero(int addr);
#ifdef ATTEMPT_IMPLEMENT_FORK
	void addFork(int addr);
#endif
	void addCall();
	void addRet();
	void addExit();
	void addNotBool();
	void addMulInt();
	void addDivInt();
	void addModInt();
	void addAddInt();
	void addSubInt();
	void addNegInt();
	void addNotInt();
	void addSHLInt();
	void addSHRInt();
	void addGInt();
	void addLInt();
	void addGEInt();
	void addLEInt();
	void addEQInt();
	void addNEInt();
	void addGFloat();
	void addLFloat();
	void addGEFloat();
	void addLEFloat();
	void addAndInt();
	void addXorInt();
	void addOrInt();
	void addMulFloat();
	void addDivFloat();
	void addAddFloat();
	void addSubFloat();
	void addNegFloat();

	int getLastOp();
	int getIndex();
	int get(int index);
	void set(int i, int val);
	void push(int i);
	
	int printOp(int data);
	void print();

	bool swapBlock(int start, int mid, int end);
	
	CCode();

	void clear();
};

typedef struct {
	const char *name;
	int param_count;
} op_t;

extern op_t ops[OP_COUNT];		//must align with the OP_ enum

// ************ primitives ************ 

enum {
	PT_INVALID = 0,
	PT_VOID,
	PT_BOOL,
	PT_INT,
	PT_FLOAT,
	PT_ENTITY,
	PT_THREAD,
	PT_STRING,
	PT_COUNT
};

//exclude PT_INVALID from PT_FIRST / PT_LAST range
#define PT_FIRST	PT_VOID
#define PT_LAST		PT_STRING

/*
universal type index table:
 x<0		function
 x=0		invalid
 0<x<N		primitive
 x>=N		structure
*/

/*
order of scope searching:
#1) local var						PUSH_REL	<relative address of variable>
#2) global var / func				PUSH_GLOB	<global address of variable>
#3) predef var / func (extern'd)	PUSH_EXT	<index of external variable>

req tables:
#1) local vars			(rw)
#2) global funcs		(r)		(same as const declared vars)
#3) global vars			(rw)
#4) predef funcs		(r)
#5) predef vars			(r)

universal var class:
	string name		<- name of the variable / function
	int type		<- universal type index (function / primitive / structure)
	int reftype		<- enum : local var / global / predef
	int ref	
	<- reftype == local var:	relative address
	<- reftype == global, type isa variable / structure:	fixed address in global stack to data
	<- reftype == global, type isa function:				index in function table
	<- reftype == predef, type isa variable / structure:	pointer to (int*)
	<- reftype == predef, type isa function:				pointer to void (*)(void)

new req cmds:
PUSH_REL	<int rel_index>		//push from thread stack relative to thread SP
PUSH_GLOB	<int glob_index>	//push from global stack
PUSH_EXT	<int ext_index>		//push from external table pointer
PUSH_CONST	<int value>			//push constant value

POP_REL							//pop into thread stack relative to thread SP
POP_GLOB						//pop into global stack
POP_EXT							//pop into external table pointer

CALL							//call 
CALL_EXT
*/

// ************ structures ************

//TODO - link this up with function instance variables
// - theres a similarity with the whole offset issue
class CStructParam {
public:
	int typeIndex;
	string name;
};

class CStruct {
public:
	string name;
	vector<CStructParam> field;	//array of types

	int size(CProgram *program);

	int getFieldIndex(const char *name);

	int getFieldOffset(CProgram *program, int fieldIndex);
};

// ************ function types ************ 

class CFuncType {
public:
	int returnType;
	int paramCount;
	int paramTypes[8];	//dyn alloc somewhere else?  save as a big lump somewhere else?

	CFuncType();
	CFuncType(int returnType, int paramCount, int *params);

	void addParam(int type);	//index ref to param table 
	void clear();
	void printNumeric();
	void print(CProgram *program);
	int getName(char *dest, int size, CProgram *program);		//returns # chars filled <excluding \0>
	int getParamSetTypeSize(CProgram *program);

	bool equals(const CFuncType *funcType);
};

// ************ variables ************ 

class CVariable {
public:
#ifdef DATA_STRING
	string name;
#else
	char name[256];
#endif

	int type;

	int size(CProgram *program);
};

// ************ predefined variables ************ 

class CPredefVariable {
public:
#ifdef DATA_STRING
	string name;
#else
	char name[256];
#endif
	int type;
	int *addr;
};

// ************ functions ************ 

class CFunction {
public:
#ifdef DATA_STRING
	string name;
#else
	char name[256];
#endif
	int functype;		//ref to funcTypeTable
	int localSize;		//size of local variables
	int codeOffset;
};


// ************ predefined functions ************ 

class CPredefFunction {
public:
#ifdef DATA_STRING
	string name;
#else
	char	name[256];
#endif
	int		functype;		//ref to funcTypeTable
	void	(*func)(CThreadStackInterface *tsinterface);
};


// ************ program - the wrapper of it all ************ 

class CProgram {
public:
	// ************ code ************ 

	CCode code;

	//mixed with primitives - unifying function types, primitive types, and structure types
	const char *getTypeName(int type);

	// ************ primitives ************

	static int getPrimTypeID(const char *str);

	// ************ structures ************

	vector<CStruct> structTable;

	int getStructByName(const char *name);
	int registerStruct(const char *name);

	// unified prim/struct type set

	//we can't support functions here because, to do so, we would have to
	//compile the string parameter passed -- and thats just not cool
	int getPrimOrStructTypeID(const char *name);

	void printStructTable(void);

	int getTypeSize(int type);

	// ************ function types ************ 

#ifdef DATA_STRING
	vector<CFuncType> funcTypeTable;
#else
	CFuncType funcTypeTable[256];
	int funcTypeTableCount;
#endif

	int registerFuncType(CFuncType funcType);
	void printFuncTypeTable(void);

//i am pretty sure this is only needed for compile time - not for run time
//in other words - remove this from CProgram and place in CCodeCompiler
	// ************ local variables ************ 

#ifdef DATA_STRING
	vector<CVariable> variableTable;
#else
	CVariable variableTable[256];
	int variableTableCount;
#endif

	int registerVariable(const char *name, int type);
	int variableTableIndex(const char *name);

#ifndef SEPARATE_GLOBALS
	//returns the size of the variable table stack.
	//used for initializing threads to make room for the global variables
	int variableTableSize(void);
#endif

	void printVariableTable(void);

#ifdef SEPARATE_GLOBALS
	// ************ global variables ************ 

	vector<CVariable> globVarTable;

	int registerGlobVar(const char *name, int type);
	int globVarTableIndex(const char *name);

	//returns the size of the variable table stack.
	//used for initializing threads to make room for the global variables
	int globVarTableSize(void);

	void printGlobVarTable(void);
#endif

	// ************ predefined variables ************ 

#ifdef DATA_STRING
	vector<CPredefVariable> predefVariableTable;
#else
	CPredefVariable predefVariableTable[256];
	int predefVariableTableCount;
#endif

	int registerPredefVariable(const char *name, int type, void *addr);
	int predefVariableTableIndex(const char *name);
	void printPredefVariableTable(void);

	// ************ functions ************ 

#ifdef DATA_STRING
	vector<CFunction> functionTable;
#else
	CFunction functionTable[256];		//what a mess
	int functionTableCount;
#endif

	int registerFunction(const char *name, int functype);
	int functionTableIndex(const char *name);
	void printFunctionTable(void);
	void beginParamVars();
	void beginFuncVars(int funcIndex);
	void endFuncVars();

	bool isVarRelative(int index);

	// ************ all variables ************ 

	int getVariableAddress(int index);

	// ************ predefined functions ************ 

#ifdef DATA_STRING
	vector<CPredefFunction> predefFunctionTable;
#else
	CPredefFunction predefFunctionTable[256];
	int predefFunctionTableCount;
#endif

	int registerPredefFunction(const char *name, int functype, void (*func)(CThreadStackInterface *));
	int predefFunctionTableIndex(const char *name);
	void printPredefFunctionTable(void);

	// ************ the unified function set ************ 
		//interface for predef and non-predef functions

	bool getFunctionID(const char *name, int *ret);
	bool getFunctionType(int index, int *ret);

	// ************ the unified variable set ************ 
		//interface for predef and non-predef variables

	bool getVariableID(const char *name, int *ret);
	int getVariableType(int index);

	// ************ strings ************ 

#ifdef DATA_STRING
	vector<string> stringTable;
#else
	string stringTable[256];
	int stringTableCount;
#endif

	int registerString(const char *name);
	int stringTableIndex(const char *name);
	void printStringTable(void);

	// ************ misc ************ 

	void print();

	void clear();

	CProgram();
};

#ifdef DATA_STRING
bool operator< (const CFuncType &a, const CFuncType &b);
bool operator== (const CFuncType &a, const CFuncType &b);
bool operator< (const CVariable &a, const CVariable &b);
bool operator== (const CVariable &a, const CVariable &b);
bool operator< (const CFunction &a, const CFunction &b);
bool operator== (const CFunction &a, const CFunction &b);
bool operator< (const CPredefVariable &a, const CPredefVariable &b);
bool operator== (const CPredefVariable &a, const CPredefVariable &b);
bool operator< (const CPredefFunction &a, const CPredefFunction &b);
bool operator== (const CPredefFunction &a, const CPredefFunction &b);
bool operator< (const CStruct &a, const CStruct &b);
bool operator== (const CStruct &a, const CStruct &b);
bool operator< (const CStructParam &a, const CStructParam &b);
bool operator== (const CStructParam &a, const CStructParam &b);

#endif

#endif