/*

    File: readfile.cpp

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <stdio.h>
#include <stdlib.h>
#include "readfile.h"

void CReadBuffer::clear(void) {
	data = NULL;
	size = 0;
	gotData = false;
}

CReadBuffer::CReadBuffer(void) {
	clear();
}

CReadBuffer::CReadBuffer(const char *filename) {
	clear();
	loadFile(filename);
}

CReadBuffer::~CReadBuffer(void) {
	if (data) free(data);
	clear();
}

bool CReadBuffer::loadFile(const char *filename) {

	clear();	//in case we already had some

	FILE *file;

	file = fopen(filename, "rb");
	if (file) {

		if (!fseek(file, 0, SEEK_END)) {

			size = ftell(file);
			if (size != -1L) {

				size+=2;	//null-term for kix.

				if (!fseek(file, 0, SEEK_SET)) {

					data = (char *)malloc(size);
					if (data) {

						if (fread(data, size-2, 1, file) == 1) {
							data[size-2] = 0;
							data[size-1] = 0;
							printf("ReadFile completed successfully\n");
						} else {
							printf("fread() failed\n");
						}

					} else {
						printf("malloc(%d) failed\n", size);
					}

				} else {
					printf("fseek() failed\n");
				}

			} else {
				printf("ftell() failed\n");
			}

		} else {
			printf("fseek() failed\n");
		}

		fclose(file);
	} else {
		printf("fopen(%s, \"rb\") failed\n", filename);
	}

	if (!data) {
		clear();
		return false;
	}

	//the only way this'll get set
	gotData = true;

	return true;
}

bool CReadBuffer::inbounds(int index) {
	return (index >= 0 && index < size);
}

char CReadBuffer::getChar(int index) {
	if (!inbounds(index)) return 0;
	return data[index];
}

bool CReadBuffer::hasData() {
	return gotData;
}

int CReadBuffer::getSize() {
	return size;
}

const char *CReadBuffer::getData() {
	return data;
}
