/*

    File: thread.cpp

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <assert.h>

//#define HACK_HACK

#ifdef HACK_HACK
#include "../m_foundation/main.h"
#endif

#include "thread.h"

#define numberof(x)		(sizeof(x)/sizeof(x[0]))

#undef VERBOSE_OUT
#undef VERBOSE_THREAD

//#define VERBOSE_THREAD
//#define VERBOSE_OUT

bool CScriptThread::reset(int funcIndex, int funcStackSize, int *funcStack) {

#if 0	//STORE_GLOBALS_IN_LOCAL_STACK

	//initialize the base pointer after all the global variables
	bp = program->variableTableSize();
	//initialize the stack pointer to just after the global variable space.
#ifdef DATA_STRING
	dataStack.resize(bp);
#else
	sp = bp;
#endif

#else

	bp = 0;

#endif

	//init the instruction pointer to the 'end of the world' call code
	ip = -1;

	//find the 'main' method - set ip to its code offset
	CFuncType *funcType = &program->funcTypeTable[
		program->functionTable[funcIndex].functype];
#ifdef DATA_STRING
	dataStack.resize(dataStack.size() + program->getTypeSize(funcType->returnType));	//make room for 'main's return

	int paramSize = funcType->getParamSetTypeSize(program);
	dataStack.resize(dataStack.size() + paramSize);				//make room for 'main's params
	if (funcStack) {			//if we are given initial stack data...
		if (funcStackSize > paramSize) funcStackSize = paramSize;	//clamp the upper bound of the initial stack data copied
		memcpy(&dataStack[dataStack.size() - paramSize], funcStack, funcStackSize * sizeof(int));	//copy it over
	}
	
#else
	sp += program->getTypeSize(funcType->returnType);	//make room for 'main's return
	sp += funcType->getParamSetTypeSize(program);				//make room for 'main's params
#endif
	return doCall(funcIndex);								//enter the main function
}

int CScriptThread::getNext() {
	int ret = program->code.get(ip);
	ip++;
	return ret;
}

bool CScriptThread::doCall(int funcIndex) {
	if (funcIndex < 0) {	//external function call
		funcIndex = ~funcIndex;
#ifdef DATA_STRING
		if (funcIndex >= program->predefFunctionTable.size()) {
#else
		if (funcIndex >= program->predefFunctionTableCount) {
#endif
#ifdef VERBOSE_THREAD
			printf("TRIED TO EXECUTE AN UNDEFINED PREDEF FUNCTION\n");
#endif
			return false;
		}
		CPredefFunction *func = &program->predefFunctionTable[funcIndex];
		//func->func(this, program);
		sfinterface.dataStack = &dataStack;
		sfinterface.program = program;
		sfinterface.func = funcIndex;
		sfinterface.globalStack = globalStack;
		func->func(&sfinterface);
	} else {		//internal function call
		push(ip);
		push(bp);
#ifdef DATA_STRING
		bp = dataStack.size();
		if (funcIndex >= program->functionTable.size()) {
#else
		bp = sp;
		if (funcIndex >= program->functionTableCount) {
#endif
#ifdef VERBOSE_THREAD
			printf("TRIED TO EXECUTE AN UNDEFINED FUNCTION\n");
#endif
			return false;
		}
		CFunction *func = &program->functionTable[funcIndex];
#ifdef DATA_STRING
		dataStack.resize(dataStack.size() + func->localSize);
#else
		sp += func->localSize;
#endif
		ip = func->codeOffset;
	}
	return true;
}

#ifdef ATTEMPT_IMPLEMENT_FORK
	//clones this thread into a new thread
	//returns the thread handle
int CScriptThread::fork(int func) {

	//issue - now we need a process class to keep track of all our threads
	//and to set their priority and what not

}
#endif

void CScriptThread::push(int data) {
#ifdef DATA_STRING
	dataStack.push_back(data);
#else
	if (sp < 0 || sp >= numberof(dataStack)) {
#ifdef VERBOSE_THREAD
		printf("stack overflow!\n");
#endif
		return;
	}
	dataStack[sp] = data;
	sp++;
#endif
}

int CScriptThread::pop(void) {
#ifdef DATA_STRING
	if (dataStack.empty()) {
#ifdef VERBOSE_THREAD
		printf("stack underflow!\n");
#endif
		return 0;
	}
	int top = dataStack[dataStack.size()-1];
	dataStack.pop_back();
	return top;
#else
	sp--;
	if (sp < 0 || sp >= numberof(dataStack)) {
#ifdef VERBOSE_THREAD
		printf("stack underflow!\n");
#endif
		return 0;
	}
	return dataStack[sp];
#endif
}

bool CScriptThread::doNext() {
	typedef union {
		int i;
		float f;
	} intfloat_t;
	intfloat_t p1, p2;

	//if we're out of bounds then fail immediately
#ifdef DATA_STRING
#else
	if (sp < 0 || sp >= numberof(dataStack)) return false;
#endif

#ifdef VERBOSE_THREAD
	program->code.printOp(ip);
#endif

	int instr = getNext();

	switch(instr) {
	case OP_PUSH_CONST:	//PUSH_CONST	<int>					- push immediate onto stack
#ifndef VERBOSE_OUT
		push(getNext());
#else
		{
			int c = getNext();
			push(c);
			printf("\t\t\t\tstack[%d] = %d ; sp = %d\n", sp-1, c, sp);
		}
#endif
		break;
	case OP_PUSH_REL:	//PUSH_REL		<addr>					- push data at stack location [BP+addr]
		push(dataStack[bp+getNext()-1]);
		break;
	case OP_PUSH_STR_REL:
		p1.i = getNext();	//rel.addr
		p2.i = getNext();	//size
		p1.i = bp + p1.i - 1;	//abs.addr
		while(p2.i) {
			push(dataStack[p1.i]);	//reference the stack forwards =)
			p1.i++;
			p2.i--;
		}
		break;
	case OP_PUSH_GLOB:	//PUSH_GLOB		<addr>					- push data at stack location [addr]
		p1.i = getNext();

#ifdef HACK_HACK
		main_console->Print("OP_PUSH_GLOB %d\n", p1.i);
#endif

		if (p1.i >= 0) {
#if 0	//STORE_GLOBALS_IN_LOCAL_STACK
			push(dataStack[p1.i]);
#else
			push(globalStack->get(p1.i));
#endif
		} else {

#ifdef HACK_HACK
			//debug stuff
			{
				CPredefVariable *v = &(program->predefVariableTable[~p1.i]);

				main_console->Print("pushing predef variable %s / %d @ %p := %d\n",
					v->name.c_str(),
					~p1.i,
					v->addr,
					v->addr[0]);

				push(v->addr[0]);
			}
#else
			push(program->predefVariableTable[~p1.i].addr[0]);
#endif
		}
		break;
	case OP_PUSH_STR_GLOB:
		p1.i = getNext();	//abs.addr
		p2.i = getNext();	//size
		if (p1.i >= 0) {
			while(p2.i) {
				push(dataStack[p1.i]);	//reference the stack forwards =)
				p1.i++;
				p2.i--;
			}
		} else {
			int i = 0;
			while(p2.i) {
				push(program->predefVariableTable[~p1.i].addr[i]);	//reference the stack forwards =)
				i++;
				p2.i--;
			}
		}
		break;
	case OP_POP_REL:		//POP_REL		<addr>					- pops data into stack location [BP+addr]
		dataStack[bp + getNext() - 1] = pop();
		break;
	case OP_POP_STR_REL:
		p1.i = getNext();	//rel.addr
		p2.i = getNext();	//size
		p1.i += bp - 1;		//abs.addr
		p1.i += p2.i - 1;	//end of space
		while(p2.i) {
			dataStack[p1.i] = pop();	//reference the stack backwards =) to compenseate for pop()'s
			p1.i--;
			p2.i--;
		}
/*
#ifdef DATA_STRING
		memcpy(
			&(dataStack[bp + p1.i]),	//relative address - BP in the dataStack with the addr offset
			&(dataStack[dataStack.size() - p2.i]),	//source - the current dataStack location, minus the size of the copied data
			p2.i * sizeof(int));
#else
		memcpy(
			&(dataStack[bp + p1.i]),	//relative address - BP in the dataStack with the addr offset
			&(dataStack[sp - p2.i]),	//source - the current dataStack location, minus the size of the copied data
			p2.i * sizeof(int));
#endif
#error shouldn't we mod dataStack.size()?
*/

		break;
	case OP_POP_GLOB:	//POP_GLOB		<addr>					- pops data into dataStack location [addr]
		p1.i = getNext();

#ifdef HACK_HACK
		main_console->Print("OP_POP_GLOB %d\n", p1.i);
#endif

		if (p1.i >= 0) {
#if 0	//STORE_GLOBALS_IN_LOCAL_STACK
			dataStack[p1.i] = pop();
#else
			globalStack->set(p1.i, pop());
#endif
		} else {

#ifdef HACK_HACK
			//debug stuff
			{
				CPredefVariable *v = &(program->predefVariableTable[~p1.i]);
				int value = pop();
				main_console->Print("popping predef variable %s / %d @ %p with %d\n",
					v->name.c_str(),
					~p1.i,
					v->addr,
					value);
				program->predefVariableTable[~p1.i].addr[0] = value;
			}
#else
			program->predefVariableTable[~p1.i].addr[0] = pop();
#endif
		}
		break;
	case OP_POP_STR_GLOB:
		p1.i = getNext();	//abs.addr
		p2.i = getNext();	//size
		if (p1.i >= 0) {
			p1.i += p2.i - 1;	//end of space
			while(p2.i) {
				dataStack[p1.i] = pop();	//reference the stack backwards =) to compenseate for pop()'s
				p1.i--;
				p2.i--;
			}
		} else {
			int i = 0;
			i += p2.i - 1;
			while(p2.i) {
				program->predefVariableTable[~p1.i].addr[i] = pop();	//reference the stack backwards =) to compenseate for pop()'s
				i--;
				p2.i--;
			}
		}
		break;
	case OP_ADD_SP:
#ifdef DATA_STRING
		{
			int oldSize = dataStack.size();
			int addSize = getNext();
			int newSize = oldSize + addSize;
			dataStack.resize(newSize);
		}
#else
		sp += getNext();

#error bullet-proof me.

#endif
#ifdef VERBOSE_OUT
		printf("\t\t\t\tsp = %d\n", sp);
#endif
		break;
	case OP_JUMP:
		ip += getNext();
		break;
	case OP_JUMP_ZERO:
		p1.i = getNext();	//addr
		if (!pop()) ip += p1.i;
		break;
#ifdef ATTEMPT_IMPLEMENT_FORK
	case OP_FORK:
		p1.i = getNext();
		//create a new thread with all the old thread parameters and priorities
		push(fork());
		ip += p1.i;
		break;
#endif
	case OP_CALL:		//CALL			<int>					- calls the function indexed by the int.  if int>=0 then calls a code function.  if int<0 then calls a predef function
		return doCall(pop());
	case OP_RET:			//RET									- returns from the last call
#ifdef DATA_STRING
		dataStack.resize(bp);
#else
		sp = bp;
#endif
		bp = pop();
		ip = pop();
		if (ip == -1) {
#ifdef VERBOSE_THREAD
			printf("done\n");
#endif
			return false;
		}
		break;
	case OP_EXIT:
		return false;	//indicates we're done.
	case OP_NOT_B:
#ifdef DATA_STRING
		dataStack[dataStack.size()-1] = !dataStack[dataStack.size()-1];
#else
		dataStack[sp-1] = !dataStack[sp-1];
#endif
		break;
	case OP_MUL_I:		//MUL_I									- pops two ints from the dataStack, multiplies them, pushs the result back
#ifdef DATA_STRING
		dataStack[dataStack.size()-2] *= dataStack[dataStack.size()-1];
		dataStack.pop_back();
#else
		dataStack[sp-2] *= dataStack[sp-1];		sp--;		break;
#endif
	case OP_DIV_I:		//DIV_I									- pops two ints from the dataStack, divides them, pushs the result back
#ifdef DATA_STRING
		if (dataStack[dataStack.size()-1]) {
			dataStack[dataStack.size()-2] /= dataStack[dataStack.size()-1];
			dataStack.pop_back();
			break;
#else
		if (dataStack[sp-1]) {
			dataStack[sp-2] /= dataStack[sp-1];		sp--;		break;
#endif
		} else {
#ifdef VERBOSE_THREAD
			printf("attempted div by zero\n");
#endif
		}
		return false;
	case OP_MOD_I:
#ifdef DATA_STRING
		if (dataStack[dataStack.size()-1]) {
			dataStack[dataStack.size()-2] %= dataStack[dataStack.size()-1];
			dataStack.pop_back();
			break;
#else
		if (dataStack[sp-1]) {
			dataStack[sp-2] %= dataStack[sp-1];		sp--;		break;
#endif
		} else {
#ifdef VERBOSE_THREAD
			printf("attempted mod by zero\n");
#endif
		}
		return false;
	case OP_ADD_I:		//ADD_I									- pops two ints from the dataStack, adds them, pushs the result back
#ifdef DATA_STRING
		dataStack[dataStack.size()-2] += dataStack[dataStack.size()-1];
		dataStack.pop_back();
		break;
#else
		dataStack[sp-2] += dataStack[sp-1];		sp--;		break;
#endif
	case OP_SUB_I:		//SUB_I									- pops two ints from the dataStack, subtracts them, pushs the result back
#ifdef DATA_STRING
		dataStack[dataStack.size()-2] -= dataStack[dataStack.size()-1];
		dataStack.pop_back();
		break;
#else
		dataStack[sp-2] -= dataStack[sp-1];		sp--;		break;
#endif
	case OP_NEG_I:
#ifdef DATA_STRING
		dataStack[dataStack.size()-1] = -dataStack[dataStack.size()-1];
#else
		dataStack[sp-1] = -dataStack[sp-1];
#endif
		break;
	case OP_NOT_I:
#ifdef DATA_STRING
		dataStack[dataStack.size()-1] = ~dataStack[dataStack.size()-1];
#else
		dataStack[sp-1] = ~dataStack[sp-1];
#endif
		break;
	case OP_SHL_I:
#ifdef DATA_STRING
		dataStack[dataStack.size()-2] <<= dataStack[dataStack.size()-1];
		dataStack.pop_back();
		break;
#else
		dataStack[sp-2] <<= dataStack[sp-1];		sp--;		break;
#endif
	case OP_SHR_I:
#ifdef DATA_STRING
		dataStack[dataStack.size()-2] >>= dataStack[dataStack.size()-1];
		dataStack.pop_back();
		break;
#else
		dataStack[sp-2] >>= dataStack[sp-1];		sp--;		break;
#endif
	case OP_G_I:

#ifdef DATA_STRING
		dataStack[dataStack.size()-2] =
			dataStack[dataStack.size()-2] > dataStack[dataStack.size()-1];
		dataStack.pop_back();
		break;
#else
		dataStack[sp-2] = dataStack[sp-2] > dataStack[sp-1];		sp--;		break;
#endif
	case OP_L_I:
#ifdef DATA_STRING
		dataStack[dataStack.size()-2] =
			dataStack[dataStack.size()-2] < dataStack[dataStack.size()-1];
		dataStack.pop_back();
		break;
#else
		dataStack[sp-2] = dataStack[sp-2] < dataStack[sp-1];		sp--;		break;
#endif
	case OP_GE_I:
#ifdef DATA_STRING
		dataStack[dataStack.size()-2] =
			dataStack[dataStack.size()-2] >= dataStack[dataStack.size()-1];
		dataStack.pop_back();
		break;
#else
		dataStack[sp-2] = dataStack[sp-2] >= dataStack[sp-1];		sp--;		break;
#endif
	case OP_LE_I:
#ifdef DATA_STRING
		dataStack[dataStack.size()-2] =
			dataStack[dataStack.size()-2] <= dataStack[dataStack.size()-1];
		dataStack.pop_back();
		break;
#else
		dataStack[sp-2] = dataStack[sp-2] <= dataStack[sp-1];		sp--;		break;
#endif
	case OP_EQ_I:
#ifdef DATA_STRING
		dataStack[dataStack.size()-2] =
			dataStack[dataStack.size()-2] == dataStack[dataStack.size()-1];
		dataStack.pop_back();
		break;
#else
		dataStack[sp-2] = dataStack[sp-2] == dataStack[sp-1];		sp--;		break;
#endif
	case OP_NE_I:
#ifdef DATA_STRING
		dataStack[dataStack.size()-2] =
			dataStack[dataStack.size()-2] != dataStack[dataStack.size()-1];
		dataStack.pop_back();
		break;
#else
		dataStack[sp-2] = dataStack[sp-2] != dataStack[sp-1];		sp--;		break;
#endif
	case OP_G_F:
#ifdef DATA_STRING
		dataStack[dataStack.size()-2] =
			*(float *)(&dataStack[dataStack.size()-2])
			> *(float *)(&dataStack[dataStack.size()-1]);
		dataStack.pop_back();
		break;
#else
		dataStack[sp-2] = *(float *)(&dataStack[sp-2]) > *(float *)(&dataStack[sp-1]);	sp--;	break;
#endif
	case OP_L_F:
#ifdef DATA_STRING
		dataStack[dataStack.size()-2] =
			*(float *)(&dataStack[dataStack.size()-2])
			< *(float *)(&dataStack[dataStack.size()-1]);
		dataStack.pop_back();
		break;
#else
		dataStack[sp-2] = *(float *)(&dataStack[sp-2]) < *(float *)(&dataStack[sp-1]);	sp--;	break;
#endif
	case OP_GE_F:
#ifdef DATA_STRING
		dataStack[dataStack.size()-2] =
			*(float *)(&dataStack[dataStack.size()-2])
			>= *(float *)(&dataStack[dataStack.size()-1]);
		dataStack.pop_back();
		break;
#else
		dataStack[sp-2] = *(float *)(&dataStack[sp-2]) >= *(float *)(&dataStack[sp-1]);	sp--;	break;
#endif
	case OP_LE_F:
#ifdef DATA_STRING
		dataStack[dataStack.size()-2] =
			*(float *)(&dataStack[dataStack.size()-2])
			<= *(float *)(&dataStack[dataStack.size()-1]);
		dataStack.pop_back();
		break;
#else
		dataStack[sp-2] = *(float *)(&dataStack[sp-2]) <= *(float *)(&dataStack[sp-1]);	sp--;	break;
#endif
	case OP_AND_I:
#ifdef DATA_STRING
		dataStack[dataStack.size()-2] &= dataStack[dataStack.size()-1];
		dataStack.pop_back();
		break;
#else
		dataStack[sp-2] &= dataStack[sp-1];		sp--;		break;
#endif
	case OP_XOR_I:
#ifdef DATA_STRING
		dataStack[dataStack.size()-2] ^= dataStack[dataStack.size()-1];
		dataStack.pop_back();
		break;
#else
		dataStack[sp-2] ^= dataStack[sp-1];		sp--;		break;
#endif
	case OP_OR_I:
#ifdef DATA_STRING
		dataStack[dataStack.size()-2] |= dataStack[dataStack.size()-1];
		dataStack.pop_back();
		break;
#else
		dataStack[sp-2] |= dataStack[sp-1];		sp--;		break;
#endif
	case OP_MUL_F:		//MUL_F									- pops two floats from the dataStack, multiplies them, pushs the result back
#ifdef DATA_STRING
		*((float *)&dataStack[dataStack.size()-2]) =
			*((float *)&dataStack[dataStack.size()-2])
			* *((float *)&dataStack[dataStack.size()-1]);
		dataStack.pop_back();
		break;
#else
		*((float *)&dataStack[sp-2]) = *((float *)&dataStack[sp-2]) * *((float *)&dataStack[sp-1]);
		sp--;
		break;
#endif
	case OP_DIV_F:		//DIV_F									- pops two floats from the dataStack, divides them, pushs the result back
#ifdef DATA_STRING
		*((float *)&dataStack[dataStack.size()-2]) =
			*((float *)&dataStack[dataStack.size()-2])
			/ *((float *)&dataStack[dataStack.size()-1]);
		dataStack.pop_back();
		break;
#else
		*((float *)&dataStack[sp-2]) = *((float *)&dataStack[sp-2]) / *((float *)&dataStack[sp-1]);
		sp--;
		break;
#endif
	case OP_ADD_F:		//ADD_F									- pops two floats from the dataStack, adds them, pushs the result back
#ifdef DATA_STRING
		*((float *)&dataStack[dataStack.size()-2]) =
			*((float *)&dataStack[dataStack.size()-2])
			+ *((float *)&dataStack[dataStack.size()-1]);
		dataStack.pop_back();
		break;
#else
		*((float *)&dataStack[sp-2]) = *((float *)&dataStack[sp-2]) + *((float *)&dataStack[sp-1]);
		sp--;
		break;
#endif
	case OP_SUB_F:		//SUB_F									- pops two floats from the dataStack, subtracts them, pushs the result back
#ifdef DATA_STRING
		*((float *)&dataStack[dataStack.size()-2]) =
			*((float *)&dataStack[dataStack.size()-2])
			- *((float *)&dataStack[dataStack.size()-1]);
		dataStack.pop_back();
		break;
#else
		*((float *)&dataStack[sp-2]) = *((float *)&dataStack[sp-2]) - *((float *)&dataStack[sp-1]);
		sp--;
		break;
#endif
	case OP_NEG_F:
#ifdef DATA_STRING
		*((float *)&dataStack[dataStack.size()-1]) = -(*((float *)&dataStack[dataStack.size()-1]));
		break;
#else
		*((float *)&dataStack[sp-1]) = -(*((float *)&dataStack[sp-1]));
		break;
#endif
	case OP_BAD:
		return false;
	default:
#ifdef VERBOSE_THREAD
		printf("got an uninterpretable instruction: %d\n", instr);
#endif
		return false;
	}
	return true;
}


bool CScriptThread::run(int funcIndex, int *returnBuffer) {
	if (!reset(funcIndex)) return false;
	while(doNext());

	if (returnBuffer) {
//		int returnOffset =
//			- 1											//pass the IP
//			- currentFuncType.getParamSetTypeSize(program)		//pass the params
//			- program->getTypeSize(currentFuncType.returnType);	//to the start of ret data
	}

	return true;
}

bool CScriptThread::runNamed(const char *funcname, int *returnBuffer) {
	int funcIndex = program->functionTableIndex(funcname);
	if (funcIndex == -1) return false;

	return run(funcIndex, returnBuffer);
}

CScriptThread::CScriptThread(CProgram *program, CGlobalStack *globalStack) {
	this->program = program;
	this->globalStack = globalStack;
}
